#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['AppDirectoryType']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from enum import IntEnum


class AppDirectoryType(IntEnum):
    local_data = 0  # local data is data that is not shared between users
    local_low_data = 1  # local data is data that is not shared between users. This folder is intended to be used by applications that require a low privilege level to operate, such as web browsers or sandboxed applications.
    roaming_data = 2  # roaming data is data that is not shared between users but might be shared and synchronised between devices of this user. On Mac OS X, automatic system backup for a small (few kbytes) files is used
    
    local_cache = 3  # Where user-specific non-essential (cached) data should be written
    local_low_cache = 4  # Where user-specific non-essential (cached) data should be written
    roaming_cache = 5  # Where user-specific non-essential (cached) data should be written. Might be shared and synchronised between devices of this user
    
    local_temp = 6
    local_low_temp = 7
    roaming_temp = 8
    
    local_log = 9
    local_low_log = 10
    roaming_log = 11
    
    local_config = 12
    local_low_config = 13
    roaming_config = 14
    
    local_runtime = 15  # Used for non-essential, user-specific data files such as sockets, named pipes, etc.
    local_low_runtime = 16  # Used for non-essential, user-specific data files such as sockets, named pipes, etc.
    roaming_runtime = 17  # Used for non-essential, user-specific data files such as sockets, named pipes, etc. Might be shared and synchronised between devices of this user

    program_files = 18  # program files is program static data that is shared between users
    program_files_common = 19  # program files common is program static data that is shared between applications and users. For example libraries 

    program_data = 20  # program data is non-static data that is shared between users
    program_cache = 21  # Where non-essential (cached) data should be written
    program_temp = 22
    program_log = 23
    program_config = 24
    program_runtime = 25  # Used for non-essential, data files such as sockets, named pipes, etc.

    user_profile_data = 28  # App data in user home directory. It is usually hidden directory
    user_profile_program_files = 29  # program files is program static data that is shared between users
    user_profile_program_files_common = 30  # program files common is program static data that is shared between applications and users. For example libraries 

    local_static_data = 31
    local_low_static_data = 32
    roaming_static_data = 33
    program_static_data = 34



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Callable, Union
from collections import deque

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class TreeStackItem:
    __slots__ = ('node', 'child')

    def __init__(self, node: Any, child: Optional[int]=None):
        self.node = node  # container
        self.child = child  # an item id in container


class DeepTreeStackItem(TreeStackItem):
    __slots__ = ('node', 'child', 'deep')

    def __init__(self, deep: int, node: Any, child: Optional[int] = None):
        super().__init__(node, child)
        self.deep: int = deep


StackItem = Union[TreeStackItem, DeepTreeStackItem, Any]


class Stack:
    def __init__(self, initial_state: Optional[StackItem]=None):
        self._stack = deque()
        if initial_state:
            self.push(initial_state)

    def push(self, state: StackItem):
        self._stack.append(state)

    def pop(self) -> StackItem:
        return self._stack.pop()

    def top(self) -> StackItem:
        return self._stack[-1]

    def __len__(self):
        return len(self._stack)


class StackUni(Stack):
    def __init__(self, initial_state: Optional[StackItem]=None, remove_on_pop: bool=True):
        super(StackUni, self).__init__(initial_state)
        self._stack = deque()
        self._stack_top = -1
        self._remove_on_pop = remove_on_pop
        if self._remove_on_pop:
            self.push = self._push_remove
            self.pop = self._pop_remove
            self.top = self._top_remove
            self._len__impl = self._len__impl_remove
        else:
            self.push = self._push_keep
            self.pop = self._pop_keep
            self.top = self._top_keep
            self._len__impl = self._len__impl_keep

        if initial_state:
            self.push(initial_state)

    def _push_remove(self, state: StackItem):
        self._stack.append(state)

    def _pop_remove(self) -> StackItem:
        return self._stack.pop()

    def _top_remove(self) -> StackItem:
        return self._stack[-1]

    def _push_keep(self, state: StackItem):
        # print('_push_keep: {}'.format(self._stack_top))
        self._stack_top += 1
        if len(self._stack) > self._stack_top:
            # we are not on a real top
            self._stack[self._stack_top] = state
        else:
            # We are on the real top - need to increase stack size
            self._stack.append(state)

    def _pop_keep(self) -> StackItem:
        # print('_pop_keep: {}'.format(self._stack_top))
        result = self._stack[self._stack_top]
        self._stack_top -= 1
        return result

    def _top_keep(self) -> StackItem:
        # print('_top_keep: {}'.format(self._stack_top))
        return self._stack[self._stack_top]

    def _len__impl_remove(self):
        return len(self._stack)

    def _len__impl_keep(self):
        return self._stack_top + 1

    def __len__(self):
        return self._len__impl()


def recursion_insureness(recursive_functor: Callable,
                         stack_based_functor: Callable,
                         on_changed_to_drop_in_replacement: Optional[Callable],
                         *args, **kwargs):
    try:
        return recursive_functor(*args, **kwargs)
    except RecursionError:
        pass

    if on_changed_to_drop_in_replacement:
        on_changed_to_drop_in_replacement()
    return stack_based_functor(*args, **kwargs)


recursion_drop_in_replacement = recursion_insureness
recursion_fallback_to_stack_based = recursion_insureness



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.performance_test_lib import PrecisePerformanceTestTracer
from cengal.data_containers.stack import *
from collections import deque

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def test_stack():
    print()
    print('Stack:')

    number_of_items = 100000
    st = Stack()

    pptt = PrecisePerformanceTestTracer(5.0)

    while pptt.iter():
        for i in range(number_of_items):
            st.push(i)
        for i in range(number_of_items):
            st.pop()

    with pptt as fast_iter:
        for j in fast_iter:
            for i in range(number_of_items):
                st.push(i)
            for i in range(number_of_items):
                st.pop()

    print('{} iter/s; {} seconds; {} iterations'.format(pptt.iter_per_time_unit * number_of_items,
                                                        pptt.time_spent,
                                                        pptt.iterations_made))


def test_deque():
    print()
    print('deque:')

    number_of_items = 100000
    st = deque()

    pptt = PrecisePerformanceTestTracer(5.0)

    while pptt.iter():
        for i in range(number_of_items):
            st.append(i)
        for i in range(number_of_items):
            st.pop()

    with pptt as fast_iter:
        for j in fast_iter:
            for i in range(number_of_items):
                st.append(i)
            for i in range(number_of_items):
                st.pop()

    print('{} iter/s; {} seconds; {} iterations'.format(pptt.iter_per_time_unit * number_of_items,
                                                        pptt.time_spent,
                                                        pptt.iterations_made))


def test_list():
    print()
    print('list:')

    number_of_items = 100000
    st = list()

    pptt = PrecisePerformanceTestTracer(5.0)

    while pptt.iter():
        for i in range(number_of_items):
            st.append(i)
        for i in range(number_of_items):
            st.pop()

    with pptt as fast_iter:
        for j in fast_iter:
            for i in range(number_of_items):
                st.append(i)
            for i in range(number_of_items):
                st.pop()

    print('{} iter/s; {} seconds; {} iterations'.format(pptt.iter_per_time_unit * number_of_items,
                                                        pptt.time_spent,
                                                        pptt.iterations_made))


def main():
    test_stack()
    test_deque()
    test_list()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import socket

from cengal.base.classes import BaseClassSettings
from cengal.code_flow_control.smart_values.versions.v_0 import ResultExistence
from cengal.data_containers.dynamic_list_of_pieces import \
    DynamicListOfPiecesDequeWithLengthControl
from cengal.data_containers.fast_fifo import FIFODequeWithLengthControl
from cengal.hardware.info.cpu.versions.v_0 import l2_cache_per_core

from .abstract import *
from .recv_buff_size_computer import RecvBuffSizeComputer

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


SET_OF_CONNECTION_ERRORS = {errno.ECONNRESET, errno.ECONNREFUSED, errno.ECONNABORTED, errno.EPIPE, errno.ESHUTDOWN}
INET_TYPE_CONNECTIONS = {socket.AF_INET, socket.AF_INET6}


try:
    BlockingIOError
except NameError:
    class BlockingIOError(OSError):
        pass

try:
    InterruptedError
except NameError:
    class InterruptedError(OSError):
        pass

try:
    ConnectionError
except NameError:
    class ConnectionError(OSError):
        pass

try:
    BrokenPipeError
except NameError:
    class BrokenPipeError(ConnectionError):
        pass

try:
    ConnectionAbortedError
except NameError:
    class ConnectionAbortedError(ConnectionError):
        pass

try:
    ConnectionRefusedError
except NameError:
    class ConnectionRefusedError(ConnectionError):
        pass

try:
    ConnectionResetError
except NameError:
    class ConnectionResetError(ConnectionError):
        pass


class InternalNotCriticalError(Exception):
    pass


MESSAGE_SIZE_LEN = 8
SERVER_ANSWER__KEYWORD_ACCEPTED = b'OK'


class SimpleNetworkError(Exception):
    pass


class ConnectionSettings(BaseClassSettings):
    def __init__(self,
                 connection_type: ConnectionType,
                 socket_address=None,
                 keyword: bytes = None,
                 socket_family=socket.AF_INET,
                 socket_type=socket.SOCK_STREAM,
                 socket_protocol=0,
                 socket_fileno=None,
                 backlog=0,
                 non_socket_connection_settings: NonSocketConnectionSettings = None):
        """
        :param connection_type: ConnectionType()
        :param socket_address: './main.server.AF_UNIX.socket', ('localhost', 8080), ('::', 50007, 0, 0), , ect.
        :param keyword: b'sdlkfj s894 saf 84ewksdhf sdf'. Can be None for a Super Server
        :param socket_family: AF_INET (the default), AF_INET6, AF_UNIX, AF_CAN or AF_RDS
        :param socket_type: SOCK_STREAM (the default), SOCK_DGRAM, SOCK_RAW or perhaps one of the other SOCK_ constants
        :param socket_protocol: in the case where the address family is AF_CAN the protocol should be one of CAN_RAW or
            CAN_BCM
        :param socket_fileno: If fileno is specified, the other arguments are ignored, causing the socket with the
            specified file descriptor to return
        """
        self.connection_type = connection_type
        self.keyword = keyword
        self.socket_address = socket_address
        self.socket_family = socket_family
        self.socket_type = socket_type
        self.socket_protocol = socket_protocol
        self.socket_fileno = socket_fileno
        self.backlog = backlog
        self.non_socket_connection_settings = non_socket_connection_settings
        self.expected_clients_with_empty_output_fifo = set()


class IOCoreMemoryManagement:
    def __init__(self):
        self.global__data_size_limit = ResultExistence(True, 2 * 1024**3)

        self.global_in__data_size_limit = ResultExistence(True, 512 * 1024**2)
        self.global_in__data_full_size = ResultExistence(True, 0)
        self.global_in__deletable_data_full_size = ResultExistence(True, 0)

        self.global_out__data_size_limit = ResultExistence(True, 512 * 1024**2)
        self.global_out__data_full_size = ResultExistence(True, 0)
        self.global_out__deletable_data_full_size = ResultExistence(True, 0)

    def link_to(self, parent):
        self.global__data_size_limit = parent.global__data_size_limit

        self.global_in__data_size_limit = parent.global_in__data_size_limit
        self.global_in__data_full_size = parent.global_in__data_full_size
        self.global_in__deletable_data_full_size = parent.global_in__deletable_data_full_size

        self.global_out__data_size_limit = parent.global_out__data_size_limit
        self.global_out__data_full_size = parent.global_out__data_full_size
        self.global_out__deletable_data_full_size = parent.global_out__deletable_data_full_size


class ASockIOCoreMemoryManagement(IOCoreMemoryManagement):
    def __init__(self):
        super(ASockIOCoreMemoryManagement, self).__init__()

        self.socket_read_fixed_buffer_size = ResultExistence(True,
                                                             int(l2_cache_per_core() / 2) or 1024 ** 2)
        # 1024**2 is the fastest fixed read buffer on my CPU.
        # Also ingeneral, it should be the half of the CPU cache per core (UPD: I don't remember why. Maybe to save other memory to instructions when we are dealing with big amount of connections).
        #
        # My CPU is Intel Core i5 3570:
        # Architecture	x86-64
        # Threads	4 threads
        # L2 cache	1 MB
        # L2 cache per core	0.25 MB/core
        # L3 cache	6 MB
        # L3 cache per core	1.5 MB/core

    def link_to(self, parent):
        super(ASockIOCoreMemoryManagement, self).link_to(parent)
        try:
            self.socket_read_fixed_buffer_size = parent.socket_read_fixed_buffer_size
        except AttributeError:
            pass


class Connection:
    def __init__(self,
                 connection_id=None,
                 connection_settings: ConnectionSettings = None,
                 connection__conn_addr: tuple = None,
                 connection_state: ConnectionState = None,
                 global_memory_management: ASockIOCoreMemoryManagement = None
                 ):
        """

        :param connection_id: ID for this connection
        :param connection__conn_addr: tuple(conn, addr) where conn is a socket, addr is an address
        :param global_memory_management: global memory management obj
        """
        self.id = connection_id
        self.connection_settings = connection_settings
        if connection__conn_addr is None:
            self.conn = ResultExistence(False, None)
            self.addr = ResultExistence(False, None)
        else:
            self.conn = ResultExistence(True, connection__conn_addr[0])
            self.addr = ResultExistence(True, connection__conn_addr[1])
        self.connection_state = connection_state

        self.addr_info = None
        self.host_names = None

        self.recv_buff_size_computer = RecvBuffSizeComputer()
        self.recv_buff_size = 0
        self.calc_new_recv_buff_size(0)

        self.should_be_closed = False  # socket should be closed immediately. For example because of IO error.
        self.ready_to_be_closed = False  # socket should be closed, after all messages had been sent to client.
        self.ready_for_deletion = False  # connection should be deleted immediately. For example because of unexpected
        #   keyword.

        self.keyword = None

        self.current_input_memoryview = None
        self.current_input_memoryview_offset = 0
        self.current_input_memoryview_nbytes = 0
        self.current_input_memoryview_diff = 0
        self.current_input_memoryview_message_nbytes = 0
        self.raw_input_from_client = DynamicListOfPiecesDequeWithLengthControl(
            external_data_length=global_memory_management.global_in__data_full_size)
        self.current_message_length = None  # length of current input message (or None, if size waw not read yet)
        self.input_from_client = FIFODequeWithLengthControl(
            external_data_full_size=global_memory_management.global_in__data_full_size)

        self.current_output_memoryview = None
        self.output_to_client = FIFODequeWithLengthControl(
            external_data_full_size=global_memory_management.global_out__data_full_size)

        self.this_is_raw_connection = False

        self.connected_expected_client_id = None
        self.connected_expected_client = None
        self.has_inline_processor = False

    def calc_new_recv_buff_size(self, last_recv_amount):
        self.recv_buff_size = self.recv_buff_size_computer.calc_new_recv_buff_size(last_recv_amount)

    def remove(self):
        self.raw_input_from_client.remove()
        self.input_from_client.remove()
        self.output_to_client.remove()

    def add_data_to_output_buffer(self, data):
        pass

    def add_data_to_input_buffer(self, data):
        pass


class InlineProcessor(InlineWorkerBase):
    __slots__ = ('client_id', 'keyword', 'socket_family', 'socket_type', 'socket_proto', 'addr_info', 'host_names',
                 'is_in_raw_mode', '__set__is_in_raw_mode', '__set__mark_socket_as_should_be_closed_immediately',
                 '__set__mark_socket_as_ready_to_be_closed', '__external_parameters_set_trigger', 'output_messages',
                 '__hold__client_id')

    def __init__(self, client_id=None, keyword: bytes = None, socket_family=None, socket_type=None, socket_proto=None,
                 addr_info=None, host_names=None, external_parameters_set_trigger: Set = None):
        """

        :param keyword: client keyword. You may check for a known keywords to act appropriately
        :param socket_family:
        :param socket_type:
        :param socket_proto:
        :param addr_info: result of socket.getaddrinfo() call
        :param host_names: result of socket.gethostbyaddr() call
        """
        super(InlineProcessor, self).__init__(client_id, keyword, socket_family, socket_type, socket_proto,
                                              addr_info, host_names, external_parameters_set_trigger)

        self.__hold__client_id = client_id
        self.__set__is_in_raw_mode = False
        self.__set__mark_socket_as_should_be_closed_immediately = False
        self.__set__mark_socket_as_ready_to_be_closed = False
        self.__external_parameters_set_trigger = external_parameters_set_trigger

    def set__is_in_raw_mode(self, is_in_raw_mode: bool):
        self.__set__is_in_raw_mode = is_in_raw_mode
        self.__external_parameters_set_trigger.add(self.__hold__client_id)

    def mark__socket_as_should_be_closed_immediately(self, mark_socket_as: bool):
        self.__set__mark_socket_as_should_be_closed_immediately = mark_socket_as
        self.__external_parameters_set_trigger.add(self.__hold__client_id)

    def mark__socket_as_ready_to_be_closed(self, mark_socket_as: bool):
        self.__set__mark_socket_as_ready_to_be_closed = mark_socket_as
        self.__external_parameters_set_trigger.add(self.__hold__client_id)

    def __getstate__(self):
        return self.client_id, self.keyword, self.socket_family, self.socket_type, self.socket_proto, self.addr_info, \
            self.host_names, self.is_in_raw_mode, self.__hold__client_id, self.__set__is_in_raw_mode, \
            self.__set__mark_socket_as_should_be_closed_immediately, self.__set__mark_socket_as_ready_to_be_closed, \
            self.__external_parameters_set_trigger, self.output_messages

    def __setstate__(self, state):
        self.client_id, self.keyword, self.socket_family, self.socket_type, self.socket_proto, self.addr_info, \
            self.host_names, self.is_in_raw_mode, self.__hold__client_id, self.__set__is_in_raw_mode, \
            self.__set__mark_socket_as_should_be_closed_immediately, self.__set__mark_socket_as_ready_to_be_closed, \
            self.__external_parameters_set_trigger, self.output_messages = state


class Client:
    def __init__(self, connection_settings: ConnectionSettings, client_id=None, connection_id=None):
        """ Dasdfd safd

        :param client_id: ID of the expected client
        :param connection_id: ID of the connection
        :param connection_settings: useful ConnectionSettings parameters are {connection_type, keyword} - for a client,
            and all - for the super server.

        """

        self.id = client_id
        self.connection_id = connection_id
        # self.__connection = None  # Нельзя! Потому что в этом случае, объект клиента станет несериализуемым
        self.__connection = None  # Можно! Сделаем сериализуемым через переопределение магических методов
        self.connection_settings = connection_settings
        self.connection_settings.check()

        self.will_use_raw_client_connection = False
        self.will_use_raw_connection_without_handshake = False
        self.this_is_unknown_client = False

        self.obj_for_inline_processing = None

    def __getstate__(self):
        # data_for_pickling = (
        #     self.id,
        #     self.connection_id,
        #     self.connection_settings,
        #     self.will_use_raw_client_connection,
        #     self.will_use_raw_connection_without_handshake,
        #     self.this_is_unknown_client,
        #     self.obj_for_inline_processing
        # )
        # return data_for_pickling
        return self.id, \
            self.connection_id, \
            self.connection_settings, \
            self.will_use_raw_client_connection, \
            self.will_use_raw_connection_without_handshake, \
            self.this_is_unknown_client, \
            self.obj_for_inline_processing

    def __setstate__(self, data_after_unpickling):
        self.id, self.connection_id, self.connection_settings, self.will_use_raw_client_connection, \
            self.will_use_raw_connection_without_handshake, self.this_is_unknown_client, \
            self.obj_for_inline_processing = data_after_unpickling

        self.__connection = None


class CheckIsRawConnection:
    def __call__(self, asock_io_core: 'ASockIOCore', connection_info: Connection) -> bool:
        """
        :param asock_io_core:
        :param connection_info:
        :return: "True" if it is RAW connection for Unknow Client. "False" otherwise.
        """
        result = False
        try:
            if connection_info.conn.result.family in {socket.AF_INET, socket.AF_INET6}:
                if connection_info.addr.result[0] not in asock_io_core.set_of_gate_addresses:
                    # If connected not from local IP address
                    result = True
        except:
            pass
        return result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from cengal.code_flow_control.none_or import none_or
from cengal.code_flow_control.smart_values.versions.v_0 import ResultExistence
from cengal.data_containers.fast_fifo import FIFODequeWithLengthControl
from typing import Optional, Tuple, List

"""
Динамический кусочный массив данных. Заполняется отдельными массивами байт (например принятых по TCP во время
ассинхронного режима работы сокета), но функционирует так, как буд-то это цельный непрерывный массив.
Дает значительный выигрыш в скорости по сравнению с обновлением встроенного контейнера list() за счет гораздо
меньшего количества аллокаций и копирований памяти.
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


# print("Just a lowly interpreted script.")


class DynamicListOfPieces:
    def __init__(self, joiner=None, on_hold_limit=None):
        self._init_param__joiner = joiner
        self._init_param__on_hold_limit = on_hold_limit

        self._joiner = none_or(joiner, b'')
        self._data = list()
        self._data_length = 0

        self._offset_limit = on_hold_limit or 1000
        self._useful_size = 0
        self._real_size = 0

    def add_piece_of_data(self, piece_of_data):
        self._data.append(piece_of_data)
        piece_of_data_len = len(piece_of_data)
        self._data_length += piece_of_data_len
        self._useful_size += 1
        self._real_size += 1

    def size(self):
        return self._data_length

    def qnt(self):
        return self._useful_size

    # @profile
    def get_data(self, size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        result_data_length = 0
        pieces_qnt = delta_range = self._real_size - self._useful_size

        if size < len(self._data[pieces_qnt]):
            result = self._data[pieces_qnt][:size]
            self._data_length -= size
            self._data[pieces_qnt] = self._data[pieces_qnt][size:]
            return result
        elif size == len(self._data[pieces_qnt]):
            result = self._data[pieces_qnt]
            self._data_length -= size
            pieces_qnt += 1
            self._useful_size -= 1
            if pieces_qnt >= self._offset_limit:
                self._real_size = self._useful_size
                self._data = self._data[pieces_qnt:]
            return result
        else:
            for piece_num in range(delta_range, self._real_size):
                piece = self._data[piece_num]
                piece_length = len(piece)
                if (result_data_length + piece_length) <= size:
                    result_data_length += piece_length
                    pieces_qnt += 1
                else:
                    break

            result_list = list()
            if size > result_data_length:
                splittable_piece = self._data[pieces_qnt]
                # if (len(splittable_piece) > 4096 * 1024**2) and (type(splittable_piece) is not memoryview):
                #     splittable_piece = memoryview(splittable_piece)
                needed = size - result_data_length

                last_current_data_part = splittable_piece[:needed]
                first_undecoded_data_part = splittable_piece[needed:]

                # result_list.append(last_current_data_part)
                self._data[pieces_qnt] = first_undecoded_data_part

                result_list = self._data[delta_range:pieces_qnt]
                result_list.append(last_current_data_part)

            else:
                result_list = self._data[delta_range:pieces_qnt]

            self._useful_size = self._real_size - pieces_qnt
            if pieces_qnt >= self._offset_limit:
                self._real_size = self._useful_size
                self._data = self._data[pieces_qnt:]

            self._data_length -= size

            # result = self._joiner.join(result_list)
            #
            # if len(result) > 4096:
            #     result = memoryview(result)
            #
            # return result

            return memoryview(self._joiner.join(result_list))

    # @profile
    def read_data(self, size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        result_data_length = 0
        pieces_qnt = delta_range = self._real_size - self._useful_size

        if size < len(self._data[pieces_qnt]):
            return self._data[pieces_qnt][:size]
        elif size == len(self._data[pieces_qnt]):
            return self._data[pieces_qnt]
        else:
            for piece_num in range(delta_range, self._real_size):
                piece = self._data[piece_num]
                piece_length = len(piece)
                if (result_data_length + piece_length) <= size:
                    result_data_length += piece_length
                    pieces_qnt += 1
                else:
                    break

            result_list = list()
            if size > result_data_length:
                splittable_piece = self._data[pieces_qnt]
                needed = size - result_data_length

                result_list = self._data[delta_range:pieces_qnt]
                result_list.append(splittable_piece[:needed])

            else:
                result_list = self._data[delta_range:pieces_qnt]

            # result = self._joiner.join(result_list)
            # return result

            return self._joiner.join(result_list)

    def remove(self):
        pass

    def __copy__(self):
        return DynamicListOfPieces(self._init_param__joiner,
                                   self._init_param__on_hold_limit)


class DynamicListOfPiecesWithLengthControl(DynamicListOfPieces):
    def __init__(self, joiner=None, on_hold_limit=None, on_hold_data_size_limit=None,
                 external_data_length: ResultExistence=None,
                 external_deletable_data_full_size: ResultExistence=None):
        super(DynamicListOfPiecesWithLengthControl, self).__init__(joiner, on_hold_limit)

        self._removed = False

        self._init_param__on_hold_data_size_limit = on_hold_data_size_limit
        self._init_param__external_data_length = external_data_length
        self._init_param__external_deletable_data_full_size = external_deletable_data_full_size

        self._on_hold_data_size_limit = on_hold_data_size_limit or 1024**2
        self._deletable_data_full_size = 0

        self._external_data_length = external_data_length or ResultExistence(False, 0)
        self._external_deletable_data_full_size = external_deletable_data_full_size or ResultExistence(False, 0)

    def add_piece_of_data(self, piece_of_data):
        self._data.append(piece_of_data)
        piece_of_data_len = len(piece_of_data)
        self._data_length += piece_of_data_len
        self._external_data_length.result += piece_of_data_len
        self._useful_size += 1
        self._real_size += 1

    # @profile
    def get_data(self, size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        result_data_length = 0
        pieces_qnt = delta_range = self._real_size - self._useful_size

        first_piece_len = len(self._data[pieces_qnt])
        if size < first_piece_len:
            result = self._data[pieces_qnt][:size]
            self._data_length -= size
            self._external_data_length.result -= size
            self._deletable_data_full_size += size
            self._external_deletable_data_full_size.result += size
            self._data[pieces_qnt] = self._data[pieces_qnt][size:]
            return result
        elif size == first_piece_len:
            result = self._data[pieces_qnt]
            self._data_length -= size
            self._external_data_length.result -= size
            self._deletable_data_full_size += size
            self._external_deletable_data_full_size.result += size
            pieces_qnt += 1
            self._useful_size -= 1
            if (pieces_qnt >= self._offset_limit) or (self._deletable_data_full_size > self._on_hold_data_size_limit):
                self._external_deletable_data_full_size.result -= self._deletable_data_full_size
                self._deletable_data_full_size = 0
                self._real_size = self._useful_size
                self._data = self._data[pieces_qnt:]
            return result
        else:
            for piece_num in range(delta_range, self._real_size):
                piece = self._data[piece_num]
                piece_length = len(piece)
                if (result_data_length + piece_length) <= size:
                    result_data_length += piece_length
                    pieces_qnt += 1
                else:
                    break

            result_list = list()
            if size > result_data_length:
                splittable_piece = self._data[pieces_qnt]
                # if (len(splittable_piece) > 4096 * 1024**2) and (type(splittable_piece) is not memoryview):
                #     splittable_piece = memoryview(splittable_piece)
                needed = size - result_data_length

                last_current_data_part = splittable_piece[:needed]
                first_undecoded_data_part = splittable_piece[needed:]

                # result_list.append(last_current_data_part)
                self._data[pieces_qnt] = first_undecoded_data_part

                result_list = self._data[delta_range:pieces_qnt]
                result_list.append(last_current_data_part)

            else:
                result_list = self._data[delta_range:pieces_qnt]

            self._deletable_data_full_size += size
            self._external_deletable_data_full_size.result += size
            self._useful_size = self._real_size - pieces_qnt
            if pieces_qnt >= self._offset_limit:
                self._external_deletable_data_full_size.result -= self._deletable_data_full_size
                self._deletable_data_full_size = 0
                self._real_size = self._useful_size
                self._data = self._data[pieces_qnt:]

            self._data_length -= size
            self._external_data_length.result -= size

            # result = self._joiner.join(result_list)
            #
            # if len(result) > 4096:
            #     result = memoryview(result)
            #
            # return result

            return memoryview(self._joiner.join(result_list))

    def remove(self):
        if not self._removed:
            self._external_data_length.result -= self._data_length
            self._external_deletable_data_full_size.result -= self._deletable_data_full_size
            self._removed = True

    def __copy__(self):
        return DynamicListOfPiecesWithLengthControl(self._init_param__joiner,
                                                    self._init_param__on_hold_limit,
                                                    self._init_param__on_hold_data_size_limit,
                                                    self._init_param__external_data_length,
                                                    self._init_param__external_deletable_data_full_size)

    def __del__(self):
        self.remove()


class DynamicListOfPiecesDequeWithLengthControl:
    def __init__(self, joiner=None, external_data_length: ResultExistence=None):
        self._removed = False

        self._init_param__joiner = joiner
        self._init_param__external_data_length = external_data_length

        self._joiner = none_or(joiner, b'')
        self._data = deque()
        self._data_length = 0

        self._real_size = 0

        self._external_data_length = external_data_length or ResultExistence(False, 0)

    def size(self):
        return self._data_length

    def qnt(self):
        return self._real_size

    def add_piece_of_data(self, piece_of_data):
        self._data.append(piece_of_data)
        piece_of_data_len = len(piece_of_data)
        self._data_length += piece_of_data_len
        self._external_data_length.result += piece_of_data_len
        self._real_size += 1

    # @profile
    def get_data(self, size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        first_piece_len = len(self._data[0])
        if size < first_piece_len:
            current_piece = memoryview(self._data[0])
            result = current_piece[:size]
            self._data[0] = current_piece[size:]
            self._data_length -= size
            self._external_data_length.result -= size
            return result
        elif size == first_piece_len:
            self._data_length -= size
            self._external_data_length.result -= size
            self._real_size -= 1
            # return memoryview(self._data.popleft())
            return self._data.popleft()
        else:
            result_data = deque()
            result_size = 0

            while self._real_size and (result_size < size):
                another_piece = self._data.popleft()
                another_piece_len = len(another_piece)
                result_data.append(another_piece)
                self._data_length -= another_piece_len
                self._external_data_length.result -= another_piece_len
                self._real_size -= 1
                result_size += another_piece_len
                if (result_size < size) and ((len(self._data[0]) + result_size) > size):
                    break

            if result_size < size:
                delta_size = size - result_size
                current_piece = memoryview(self._data[0])
                result_data.append(current_piece[:delta_size])
                self._data[0] = current_piece[delta_size:]
                self._data_length -= delta_size
                self._external_data_length.result -= delta_size

            # return memoryview(self._joiner.join(result_data))
            return self._joiner.join(result_data)

    def read_data(self, size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        first_piece_len = len(self._data[0])
        if size < first_piece_len:
            current_piece = memoryview(self._data[0])
            return current_piece[:size]
        elif size == first_piece_len:
            return memoryview(self._data[0])
        else:
            result_data = deque()
            result_size = 0

            index = 0
            while result_size < size:
                another_piece = self._data[index]
                index += 1
                another_piece_len = len(another_piece)
                result_data.append(another_piece)
                result_size += another_piece_len
                if (result_size < size) and ((len(self._data[index]) + result_size) > size):
                    break

            if result_size < size:
                delta_size = size - result_size
                current_piece = memoryview(self._data[index])
                result_data.append(current_piece[:delta_size])

            return self._joiner.join(result_data)

    # @profile
    def get_data_pieces(self, size) -> List[bytes]:
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return [self._joiner]

        first_piece_len = len(self._data[0])
        if size <= first_piece_len:
            self._data_length -= first_piece_len
            self._external_data_length.result -= first_piece_len
            self._real_size -= 1
            return [self._data.popleft()]
        else:
            result_data = deque()
            result_size = 0

            while self._real_size and (result_size < size):
                another_piece = self._data.popleft()
                another_piece_len = len(another_piece)
                result_data.append(another_piece)
                self._data_length -= another_piece_len
                self._external_data_length.result -= another_piece_len
                self._real_size -= 1
                result_size += another_piece_len

            return result_data

    def read_data_pieces(self, size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return [self._joiner]

        first_piece_len = len(self._data[0])
        if size <= first_piece_len:
            return [memoryview(self._data[0])]
        else:
            result_data = deque()
            result_size = 0

            index = 0
            while result_size < size:
                another_piece = memoryview(self._data[index])
                index += 1
                another_piece_len = len(another_piece)
                result_data.append(another_piece)
                result_size += another_piece_len

            return result_data

    def get_data_at_least(self, size):
        return self._joiner.join(self.get_data_pieces(size))

    def read_data_at_least(self, size):
        return self._joiner.join(self.read_data_pieces(size))

    def get_data_nearly(self, size):
        return self._joiner.join(self.get_data_pieces(min(size, self.size())))

    def read_data_nearly(self, size):
        return self._joiner.join(self.read_data_pieces(min(size, self.size())))
    
    def item_by_index(self, index: int) -> Tuple[int, int]:
        if index >= self._data_length:
            return -1, -1
        
        offset = 0
        for item_index, item in enumerate(self._data):
            item_len = len(item)
            offset += item_len
            if index < offset:
                index_within_item = item_len - (offset - index)
                return item_index, index_within_item
    
    def startswith(self, 
        __prefix,
        __start = None,
        __end = None
    ) -> bool:
        if not self._data_length:
            return False
            
        if isinstance(__prefix, tuple):
            prefix_len = max((len(item) for item in __prefix))
        else:
            prefix_len = len(__prefix)
        
        start_index = __start or 0
        prefix_end_index = start_index + prefix_len
        if __end is None:
            end_index = prefix_end_index
        else:
            end_index = __end
        
        end_index = min(end_index, prefix_end_index)
        if end_index < start_index:
            return False

        start_item_index, index_within_start_item = self.item_by_index(start_index)
        if -1 == start_item_index:
            return False
        
        end_item_index, index_within_end_item = self.item_by_index(end_index)
        if -1 == end_item_index:
            end_item_index = self._real_size -1
            index_within_end_item = len(self._data[end_item_index]) - 1
        
        return self._joiner.join(self._data[start_item_index: end_item_index]).startswith(
            __prefix, index_within_start_item, index_within_end_item)

    def find(
            self, __sub, __start = None, __end = None
        ) -> int:
        if not self._data_length:
            return False

        start_index = __start or 0
        end_index = __end or (self._data_length - 1)

        start_item_index, index_within_start_item = self.item_by_index(start_index)
        if -1 == start_item_index:
            return False
        
        end_item_index, index_within_end_item = self.item_by_index(end_index)
        if -1 == end_item_index:
            end_item_index = self._real_size -1
            index_within_end_item = len(self._data[end_item_index]) - 1
        
        return self._joiner.join(self._data[start_item_index: end_item_index]).find(
            __sub, index_within_start_item, index_within_end_item)

    def clear(self):
        self.get_data_pieces(self._real_size)

    def remove(self):
        if not self._removed:
            self._external_data_length.result -= self._data_length
            self._removed = True

    def __copy__(self):
        return DynamicListOfPiecesDequeWithLengthControl(
            self._init_param__joiner, self._init_param__external_data_length)

    def __del__(self):
        self.remove()


# TODO: implement in Cython:
class DynamicListOfPiecesMemoryviewDequeWithLengthControl:
    def __init__(self, joiner=None, external_data_length: Optional[ResultExistence] = None, mem_buffer_size: Optional[ResultExistence] = None, output_fifo: Optional[FIFODequeWithLengthControl] = None):
        self._removed = False

        self._init_param__joiner = joiner
        self._init_param__external_data_length = external_data_length
        self._init_param__mem_buffer_size = mem_buffer_size

        self._joiner = none_or(joiner, b'')
        self._data = deque()
        self._data_length = 0

        self._real_size = 0

        self._external_data_length = external_data_length or ResultExistence(False, 0)

        self.mem_buffer_size = mem_buffer_size or ResultExistence(False, 1024)
        self.mem_buffer = memoryview(bytearray(self.mem_buffer_size.result))
        self._external_data_length.result += self.mem_buffer_size.result

        self.mem_buffer_offset = 0
        self.mem_buffer_nbytes = 0
        self.mem_buffer_diff = 0
        self.mem_buffer_message_nbytes = 0

        self.output_data = output_fifo or FIFODequeWithLengthControl(self._external_data_length)

    def _allocate_new_mem_buffer(self):
        self._external_data_length.result -= len(self.mem_buffer.obj)
        self.mem_buffer = memoryview(bytearray(self.mem_buffer_size.result))
        self._external_data_length.result += self.mem_buffer_size.result

    def size(self):
        return self._data_length

    def qnt(self):
        return self._real_size

    def add_piece_of_data(self, piece_of_data):
        raise NotImplemented()

    def _add_piece_of_data(self, piece_of_data):
        self._data.append(piece_of_data)
        piece_of_data_len = len(piece_of_data)
        self._data_length += piece_of_data_len
        self._external_data_length.result += piece_of_data_len
        self._real_size += 1

    def piece_of_data_had_been_added(self, nbytes):
        self.mem_buffer = self.mem_buffer[nbytes:]
        self.mem_buffer_nbytes += nbytes
        self.mem_buffer_diff += nbytes

        if not self.mem_buffer:
            if self.mem_buffer_diff:
                self._add_piece_of_data(memoryview(self.mem_buffer.obj)[self.mem_buffer_offset:])
            self._allocate_new_mem_buffer()

    # @profile
    def get_data(self, size):
        """
        :param size:
        :return: data or None (if size > self._data_length)
        """
        if size > self._data_length:
            return None
        elif size <= 0:
            # return self._joiner.join(list())  # this is 1.6 times more faster than copy.copy(self._joiner)
            return self._joiner

        first_piece_len = len(self._data[0])
        if size < first_piece_len:
            current_piece = memoryview(self._data[0])
            result = current_piece[:size]
            self._data[0] = current_piece[size:]
            self._data_length -= size
            self._external_data_length.result -= size
            return result
        elif size == first_piece_len:
            self._data_length -= size
            self._external_data_length.result -= size
            self._real_size -= 1
            # return memoryview(self._data.popleft())
            return self._data.popleft()
        else:
            result_data = deque()
            result_size = 0

            while self._real_size and (result_size < size):
                another_piece = self._data.popleft()
                another_piece_len = len(another_piece)
                result_data.append(another_piece)
                self._data_length -= another_piece_len
                self._external_data_length.result -= another_piece_len
                self._real_size -= 1
                result_size += another_piece_len
                if (result_size < size) and ((len(self._data[0]) + result_size) > size):
                    break

            if result_size < size:
                delta_size = size - result_size
                current_piece = memoryview(self._data[0])
                result_data.append(current_piece[:delta_size])
                self._data[0] = current_piece[delta_size:]
                self._data_length -= delta_size
                self._external_data_length.result -= delta_size

            # return memoryview(self._joiner.join(result_data))
            return self._joiner.join(result_data)

    def read_data(self, size):
        raise NotImplemented()

    def remove(self):
        if not self._removed:
            self._external_data_length.result -= self._data_length
            self._external_data_length.result -= len(self.mem_buffer.obj)
            self._removed = True

    def __copy__(self):
        return DynamicListOfPiecesDequeWithLengthControl(
            self._init_param__joiner, self._init_param__external_data_length)

    def __del__(self):
        self.remove()


def get_num_pieces_with_full_size_lesser_than(iter_object, full_size):
    index = -1
    if iter_object:
        result_size = 0
        index = 0
        for piece in iter_object:
            piece_len = len(piece)
            if (result_size + piece_len) > full_size:
                break
            else:
                result_size += piece_len
                index += 1
    return index


def get_num_pieces_with_full_size_lesser_than_with_offset(iter_object, full_size, offset):
    pieces_qnt = -1
    if iter_object:
        object_size = len(iter_object)
        result_data_length = 0
        pieces_qnt = 0
        for piece_num in range(offset, object_size):
            piece = iter_object[piece_num]
            piece_length = len(piece)
            if (result_data_length + piece_length) > full_size:
                break
            else:
                result_data_length += piece_length
                pieces_qnt += 1
    return pieces_qnt



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Union, Set
from cengal.bulk_pip_actions.install import *
from cengal.system import IS_INSIDE_OR_FOR_WEB_BROWSER, OS_TYPE, KIVY_PLATFORM, KIVY_TARGET_PLATFORM
from cengal.hardware.info.cpu import cpu_info


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class ModulesLists(object):
    def __init__(self):
        self.list_type = None

        self.python2 = list()
        self.python3 = list()
        self.cpython2 = list()
        self.cpython3 = list()
        self.pypy2 = list()
        self.pypy3 = list()
        self.universal = list()

        self.linux_allowed: Set[str] = set()
        self.linux_forbidden: Set[str] = set()
        self.windows_allowed: Set[str] = set()
        self.windows_forbidden: Set[str] = set()
        self.osx_allowed: Set[str] = set()
        self.osx_forbidden: Set[str] = set()
        self.emscripten_allowed: Set[str] = set()
        self.emscripten_forbidden: Set[str] = set()

        self.kivy_linux_allowed: Set[str] = set()
        self.kivy_linux_forbidden: Set[str] = set()
        self.kivy_win_allowed: Set[str] = set()
        self.kivy_win_forbidden: Set[str] = set()
        self.kivy_macosx_allowed: Set[str] = set()
        self.kivy_macosx_forbidden: Set[str] = set()
        self.kivy_ios_allowed: Set[str] = set()
        self.kivy_ios_forbidden: Set[str] = set()
        self.kivy_android_allowed: Set[str] = set()
        self.kivy_android_forbidden: Set[str] = set()
        self.kivy_unknown_allowed: Set[str] = set()
        self.kivy_unknown_forbidden: Set[str] = set()

        self.kivy_target_linux_allowed: Set[str] = set()
        self.kivy_target_linux_forbidden: Set[str] = set()
        self.kivy_target_win_allowed: Set[str] = set()
        self.kivy_target_win_forbidden: Set[str] = set()
        self.kivy_target_macosx_allowed: Set[str] = set()
        self.kivy_target_macosx_forbidden: Set[str] = set()
        self.kivy_target_ios_allowed: Set[str] = set()
        self.kivy_target_ios_forbidden: Set[str] = set()
        self.kivy_target_android_allowed: Set[str] = set()
        self.kivy_target_android_forbidden: Set[str] = set()
        self.kivy_target_unknown_allowed: Set[str] = set()
        self.kivy_target_unknown_forbidden: Set[str] = set()

        self.arch__x86__allowed: Set[str] = set()
        self.arch__x86__forbidden: Set[str] = set()
        self.arch__x86_64__allowed: Set[str] = set()
        self.arch__x86_64__forbidden: Set[str] = set()
        self.arch__x86_32__allowed: Set[str] = set()
        self.arch__x86_32__forbidden: Set[str] = set()
        self.arch__ARM__allowed: Set[str] = set()
        self.arch__ARM__forbidden: Set[str] = set()
        self.arch__ARM_8__allowed: Set[str] = set()
        self.arch__ARM_8__forbidden: Set[str] = set()
        self.arch__ARM_8_64__allowed: Set[str] = set()
        self.arch__ARM_8_64__forbidden: Set[str] = set()
        self.arch__ARM_8_32__allowed: Set[str] = set()
        self.arch__ARM_8_32__forbidden: Set[str] = set()
        self.arch__ARM_7__allowed: Set[str] = set()
        self.arch__ARM_7__forbidden: Set[str] = set()

        self._allowed: Set[str] = set()
        self._forbidden: Set[str] = set()

    def chosen_packages(self):
        if IS_INSIDE_OR_FOR_WEB_BROWSER:
            # Must be before other systems!
            self._allowed.update(self.emscripten_allowed)
            self._forbidden.update(self.emscripten_forbidden)
        elif 'Linux' == OS_TYPE:
            self._allowed.update(self.linux_allowed)
            self._forbidden.update(self.linux_forbidden)
        elif 'Windows' == OS_TYPE:
            self._allowed.update(self.windows_allowed)
            self._forbidden.update(self.windows_forbidden)
        elif 'Darwin' == OS_TYPE:
            self._allowed.update(self.osx_allowed)
            self._forbidden.update(self.osx_forbidden)

        if 'linux' == KIVY_PLATFORM:
            self._allowed.update(self.kivy_linux_allowed)
            self._forbidden.update(self.kivy_linux_forbidden)
        elif 'win' == KIVY_PLATFORM:
            self._allowed.update(self.kivy_win_allowed)
            self._forbidden.update(self.kivy_win_forbidden)
        elif 'macosx' == KIVY_PLATFORM:
            self._allowed.update(self.kivy_macosx_allowed)
            self._forbidden.update(self.kivy_macosx_forbidden)
        elif 'ios' == KIVY_PLATFORM:
            self._allowed.update(self.kivy_ios_allowed)
            self._forbidden.update(self.kivy_ios_forbidden)
        elif 'android' == KIVY_PLATFORM:
            self._allowed.update(self.kivy_android_allowed)
            self._forbidden.update(self.kivy_android_forbidden)
        elif 'unknown' == KIVY_PLATFORM:
            self._allowed.update(self.kivy_unknown_allowed)
            self._forbidden.update(self.kivy_unknown_forbidden)

        if 'linux' == KIVY_TARGET_PLATFORM:
            self._allowed.update(self.kivy_target_linux_allowed)
            self._forbidden.update(self.kivy_target_linux_forbidden)
        elif 'win' == KIVY_TARGET_PLATFORM:
            self._allowed.update(self.kivy_target_win_allowed)
            self._forbidden.update(self.kivy_target_win_forbidden)
        elif 'macosx' == KIVY_TARGET_PLATFORM:
            self._allowed.update(self.kivy_target_macosx_allowed)
            self._forbidden.update(self.kivy_target_macosx_forbidden)
        elif 'ios' == KIVY_TARGET_PLATFORM:
            self._allowed.update(self.kivy_target_ios_allowed)
            self._forbidden.update(self.kivy_target_ios_forbidden)
        elif 'android' == KIVY_TARGET_PLATFORM:
            self._allowed.update(self.kivy_target_android_allowed)
            self._forbidden.update(self.kivy_target_android_forbidden)
        elif 'unknown' == KIVY_TARGET_PLATFORM:
            self._allowed.update(self.kivy_target_unknown_allowed)
            self._forbidden.update(self.kivy_target_unknown_forbidden)

        arch = cpu_info().arch.casefold()
        if cpu_info().is_x86:
            self._allowed.update(self.arch__x86__allowed)
            self._forbidden.update(self.arch__x86__forbidden)
            if 'x86_64'.casefold() == arch:
                self._allowed.update(self.arch__x86_64__allowed)
                self._forbidden.update(self.arch__x86_64__forbidden)
            elif 'x86_32'.casefold() == arch:
                self._allowed.update(self.arch__x86_32__allowed)
                self._forbidden.update(self.arch__x86_32__forbidden)
        
        if cpu_info().is_arm:
            self._allowed.update(self.arch__ARM__allowed)
            self._forbidden.update(self.arch__ARM__forbidden)
            if 'ARM_8'.casefold() == arch:
                self._allowed.update(self.arch__ARM_8__allowed)
                self._forbidden.update(self.arch__ARM_8__forbidden)
            elif ('ARM_8'.casefold() == arch) and (64 == cpu_info().bits):
                self._allowed.update(self.arch__ARM_8_64__allowed)
                self._forbidden.update(self.arch__ARM_8_64__forbidden)
            elif ('ARM_8'.casefold() == arch) and (32 == cpu_info().bits):
                self._allowed.update(self.arch__ARM_8_32__allowed)
                self._forbidden.update(self.arch__ARM_8_32__forbidden)
            elif 'ARM_7'.casefold() == arch:
                self._allowed.update(self.arch__ARM_7__allowed)
                self._forbidden.update(self.arch__ARM_7__forbidden)

        modules = list()
        modules += self.universal
        if '2' == PYTHON_VERSION[0]:
            if 'CPython' == PLATFORM_NAME:
                modules += self.cpython2
            elif 'PyPy' == PLATFORM_NAME:
                modules += self.pypy2
            modules += self.python2
        if '3' == PYTHON_VERSION[0]:
            if 'CPython' == PLATFORM_NAME:
                modules += self.cpython3
            elif 'PyPy' == PLATFORM_NAME:
                modules += self.pypy3
            modules += self.python3

        modules.extend(self._allowed)

        if self._forbidden:
            new_modules = list()
            for item in modules:
                if item not in self._forbidden:
                    new_modules.append(item)
            modules = new_modules

        return modules

    def bulk_install(self):
        modules = self.chosen_packages()
        if modules:
            print('INSTALLING: {}'.format(modules))
            print()

            for item in modules:
                install(item, self.list_type)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['WindowManagerCS', 'CompositorCS', 'TerminalApplication', 'TermApp']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import put_coro_to
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_loop import ShutdownLoop
from cengal.parallel_execution.coroutines.coro_tools.prepare_loop import prepare_loop
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.data_manipulation.conversion.reinterpret_cast import reinterpret_cast

from pytermgui import WindowManager as PTGWindowManager
from pytermgui.window_manager import Compositor as PTGCompositor

from typing import Any, Dict, Tuple, Optional, Callable


class CompositorCS(PTGCompositor):
    def init_cs(self, setup_coro_worker: AnyWorker, *args, **kwargs):
        self.setup_coro_worker: AnyWorker = setup_coro_worker
        self.setup_coro_worker_args: Tuple = args
        self.setup_coro_worker_kwargs: Dict = kwargs
        self.cs: CoroSchedulerType = None
        self.setup_coro_worker_value_holder: ValueExistence = None

    def ensure_cs(self):
        if self.cs is None:
            self.cs, self.setup_coro_worker_value_holder = prepare_loop(self.setup_coro_worker, *self.setup_coro_worker_args, **self.setup_coro_worker_kwargs)

    def iter_cs(self):
        in_work = self.cs.iteration()

    def draw(self, force: bool = False) -> None:
        self.ensure_cs()
        self.iter_cs()
        super().draw(force)
    
    def put_coro(self, coro_worker: AnyWorker, *args, **kwargs):
        put_coro_to((self.cs, None, False), coro_worker, *args, **kwargs)

    def stop(self) -> None:
        async def coro(i: Interface, self):
            def shutdown_handler():
                self._is_running = False
            
            i._loop.on_destroyed_handlers.add(shutdown_handler)
            await i(ShutdownLoop)
        
        self.put_coro(coro, self)


class WindowManagerCS(PTGWindowManager):
    @staticmethod
    def replace(window_manager: PTGWindowManager, model, view_setup: Callable, controller_coro_setup_worker: Optional[AnyWorker] = None, *args, **kwargs) -> 'WindowManagerCS':
        reinterpret_cast(WindowManagerCS, window_manager)
        view_setup(window_manager, model)
        window_manager.init_cs(controller_coro_setup_worker, model, *args, **kwargs)
        return window_manager

    def init_cs(self, setup_coro_worker: AnyWorker, *args, **kwargs):
        reinterpret_cast(CompositorCS, self.compositor)
        self.cs: CoroSchedulerType = None
        self.compositor.init_cs(setup_coro_worker, *args, **kwargs)
    
    def put_coro(self, coro_worker: AnyWorker, *args, **kwargs):
        self.compositor.put_coro(coro_worker, *args, **kwargs)


class TerminalApplication:
    def __init__(self):
        self.model = self.model_setup()
        self.manager: WindowManagerCS = None
    
    def run(self):
        with PTGWindowManager() as manager:
            self.manager = manager
            WindowManagerCS.replace(manager, self.model, self.ui_setup, self.controller_loop_setup)
    
    def __call__(self):
        self.run()
    
    def model_setup(self):
        raise NotImplementedError()
    
    def ui_setup(self, manager: WindowManagerCS, model):
        raise NotImplementedError()
    
    def controller_loop_setup(self, i: Interface, model):
        raise NotImplementedError()


TermApp = TerminalApplication



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['FastAggregatorRequest', 'FastAggregatorWaitFor', 'FastAggregatorPutSingle', 'FastAggregatorPutMultiple', 'FastAggregator', 'FastAggregatorWriter', 'FastAggregatorClient']


from collections import deque

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.introspection.inspect import get_exception
from typing import Hashable, Tuple, Dict, List, Any, Optional, Set, Deque


class FastAggregatorWaitFor(TypedServiceRequest[Any]):
    def __call__(self, key: Hashable, queued: bool = True) -> 'FastAggregatorWaitFor':
        """Will block coroutine untill result is ready

        Args:
            key (Hashable): _description_

        Returns:
            ServiceRequest: _description_
        """        
        return self._save(self.default__request__type__, key, queued)


class FastAggregatorPutSingle(TypedServiceRequest[None]):
    default__request__type__ = 1

    def __call__(self, key: Hashable, data: Any) -> 'FastAggregatorPutSingle':
        return self._save(self.default__request__type__, key, data)


class FastAggregatorPutMultiple(TypedServiceRequest[None]):
    default__request__type__ = 2
    
    def __call__(self, data: Dict[Hashable, List[Any]]) -> 'FastAggregatorPutMultiple':
        return self._save(self.default__request__type__, data)


class FastAggregatorRequest(ServiceRequest):
    def wait(self, key: Hashable, queued: bool = True) -> FastAggregatorWaitFor:
        """Will block coroutine untill result is ready

        Args:
            key (Hashable): _description_
            queued (bool, optional): When True - works like loadbalansers works. Otherwise gets same link to data as a first requester in a queue. Defaults to True.

        Returns:
            Any: _description_
        """        
        return FastAggregatorWaitFor[None]()(key, queued)

    def put_single(self, key: Hashable, data: Any) -> FastAggregatorPutSingle:
        return FastAggregatorPutSingle[None]()(key, data)

    def put_multiple(self, data: Dict[Hashable, List[Any]]) -> FastAggregatorPutMultiple:
        return FastAggregatorPutMultiple[None]()(data)


class FastAggregator(DualImmediateProcessingServiceMixin, TypedService[Any]):
    def __init__(self, loop: CoroSchedulerType):
        super().__init__(loop)
        self.data: Dict[Hashable, List[Any]] = dict()
        self.requester_by_key: Dict[Hashable, Deque[Set[CoroID]]] = dict()
        self._request_workers = {
            0: self._on_wait,
            1: self._on_put_single,
            2: self._on_put_multiple,
        }

    def single_task_registration_or_immediate_processing_single(
            self, key: Hashable) -> Tuple[bool, None, None]:
        exception = None
        result = None
        try:
            result = self.data[key]
            del self.data[key]
        except:
            exception = get_exception()
        
        return True, result, exception
    
    def _on_wait(self, key: Hashable, queued: bool = True) -> Tuple[bool, None, None]:
        if key not in self.requester_by_key:
            self.requester_by_key[key] = deque()
        
        if (not queued) and (self.requester_by_key[key]):
            self.requester_by_key[key][0].add(self.current_caller_coro_info.coro_id)
        else:
            self.requester_by_key[key].append({self.current_caller_coro_info.coro_id})

        self.make_live()
        return False, None, None
    
    def _on_put_single(self, key: Hashable, data: Any) -> Tuple[bool, None, None]:
        self.put_single(key, data)
        return True, None, None
    
    def _on_put_multiple(self, data: Dict[Hashable, List[Any]]) -> Tuple[bool, None, None]:
        self.put_multiple(data)
        return True, None, None
    
    def put_single(self, key: Hashable, data: Any):
        if key not in self.data:
            self.data[key] = list()

        if (key in self.requester_by_key) and self.requester_by_key[key]:
            self.make_live()

        self.data[key].append(data)
    
    def put_multiple(self, data: Dict[Hashable, List[Any]]):
        need_to_make_live = False
        
        for key in data.keys():
            if key not in self.data:
                self.data[key] = list()

            if (key in self.requester_by_key) and self.requester_by_key[key]:
                need_to_make_live = True
            
            self.data[key].extend(data[key])
        
        if need_to_make_live:
            self.make_live()

    def full_processing_iteration(self):
        requester_by_key_bak = self.requester_by_key
        possible_keys = set(requester_by_key_bak.keys()) & set(self.data.keys())
        for key in possible_keys:
            key_requesters_coro_ids_bak = requester_by_key_bak[key]
            if key_requesters_coro_ids_bak:
                first_requesters_coro_ids: Set[CoroID] = key_requesters_coro_ids_bak.popleft()
                data = self.data[key]
                for first_requester_coro_id in first_requesters_coro_ids:
                    self.register_response(first_requester_coro_id, data, None)
                
                del self.data[key]
            
            if not key_requesters_coro_ids_bak:
                del requester_by_key_bak[key]
        
        self.make_dead()

    def in_work(self) -> bool:
        return self.thrifty_in_work(bool(self.requester_by_key))


FastAggregatorWaitFor.default_service_type = FastAggregator
FastAggregatorPutSingle.default_service_type = FastAggregator
FastAggregatorPutMultiple.default_service_type = FastAggregator


class FastAggregatorWriter:
    def __init__(self) -> None:
        self.server:Optional[FastAggregator] = None
        self.data_buffer: Dict[Hashable, List[Any]] = dict()
    
    def try_find_server(self):
        loop = CoroScheduler.current_loop()
        if loop is not None:
            self.server = loop.get_service_instance(FastAggregator)
    
    def __call__(self, key: Hashable, data: Any) -> Any:
        if self.server is None:
            self.try_find_server()
        
        if self.server is None:
            if key not in self.data_buffer:
                self.data_buffer[key] = list()
            
            self.data_buffer[key].append(data)
        else:
            if self.data_buffer:
                self.server.put_multiple(self.data_buffer)
            
            self.server.put_single(key, data)


FastAggregatorClient = FastAggregatorWriter



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['reinterpret_cast_attempt', 'reinterpret_cast', 'temporary_reinterpret_cast', 'TemporaryReinterpretCast']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from contextlib import contextmanager
from typing import Type, Any, Generator, TypeVar


T = TypeVar('T')


def reinterpret_cast_attempt(new_type: Type[T], value: Any) -> bool:
    result: bool = True
    try:
        value.__class__ = new_type
    except TypeError:
        result = False
    
    return result


def reinterpret_cast(new_type: Type[T], value: Any) -> T:
    value.__class__ = new_type
    return value


@contextmanager
def temporary_reinterpret_cast(temporary_type: Type[T], value: Any) -> Generator[T, None, None]:
    original_class = value.__class__
    value.__class__ = temporary_type
    try:
        yield value
    finally:
        value.__class__ = original_class


class TemporaryReinterpretCast:
    def __init__(self, temporary_type: Type[T], value: Any):
        self.temporary_type: Type[T] = temporary_type
        self.value: Any = value
        self.original_class = value.__class__
    
    def __enter__(self):
        value = self.value
        value.__class__ = self.temporary_type
        return value
    
    def __exit__(self, type, value, traceback):
        self.value.__class__ = self.original_class



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.time_management.repeat_for_a_time import BaseTracer, GreedyTracer
from cengal.time_management.cpu_clock_cycles import perf_counter
from enum import Enum
from typing import Tuple

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class ViewIndicator:
    def __init__(self):
        self._view_set = ('-', '\\', '|', '/')
        self._view_set_size = len(self._view_set)
        self._counter = 0

    def __call__(self):
        result = self._view_set[self._counter]
        self._counter += 1
        if self._counter >= self._view_set_size:
            self._counter = 0
        return result

    @property
    def view_set(self):
        return self._view_set

    @view_set.setter
    def view_set(self, view_set: Tuple):
        self._view_set = view_set
        self._view_set_size = len(self._view_set)
        self._counter = 0


class ProgressOutType(Enum):
    none = 0
    short = 1
    full = 2


class ProgressMeter:
    """
    Example of use:
        pm = ProgressMeter(0.1, 'files')
        while True:
            ...
            pm(ProgressOutType.short)
        pm(ProgressOutType.full, force_print=True, erasable=False)

    """

    def __init__(self, update_time: float=1.0, _items_readable_name: str='iterations'
                 , default_indicator: ViewIndicator=None
                 , default_tracer: BaseTracer=GreedyTracer):
        self._update_time = update_time
        self._items_readable_name = _items_readable_name
        self._last_iterations_per_second = 0.0
        self.template_short = '  {indicator}'
        self.template_full = '  {indicator} ## {items_count} ## {items_readable_name} per second: ' \
                             '{iterations_per_second}'
        self._view_indicator = default_indicator or ViewIndicator()
        self._default_tracer = default_tracer
        self._tracer = self._default_tracer(self._update_time)
        self._iterations = 0
        self._start_time = 0.0
        self._last_tracked_time = 0.0
        self._is_first_call = True

    def __call__(self, output_type: ProgressOutType=ProgressOutType.full,
                 force_print: bool=False, erasable: bool=True) -> bool:
        self._iterations += 1
        if self._is_first_call:
            self._is_first_call = False
            self._start_time = perf_counter()
        result = not self._tracer.iter()
        if result or force_print:
            self._last_tracked_time = perf_counter()
            time_delta = self._last_tracked_time - self._start_time
            if time_delta != 0.0:
                self._last_iterations_per_second = self._iterations / time_delta
            self.print(output_type, erasable)
            self._tracer = self._default_tracer(self._update_time)
        return result
    
    def print(self, output_type: ProgressOutType=ProgressOutType.full, erasable: bool=True):
        out_message = None
        if ProgressOutType.full == output_type:
            out_message = self.template_full.format(indicator=self._view_indicator(),
                                                    items_count=self._iterations,
                                                    items_readable_name=self._items_readable_name,
                                                    iterations_per_second=self._last_iterations_per_second)
        elif ProgressOutType.short == output_type:
            out_message = self.template_short.format(indicator=self._view_indicator())

        if out_message is not None:
            if erasable:
                print(out_message, end='\r', flush=True)
            else:
                print(out_message, flush=True)

    @property
    def iterations_per_second(self):
        return self._last_iterations_per_second



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# __all__ = ['frame', 'get_exception', 'get_exception_tripple', 'exception_to_printable_text', 'is_async', 'is_callable', 'func_param_names', 'frame_param_names', 'intro_func_param_names', 'CodeParamsWithValues', 'intro_func_params_with_values', 'intro_func_all_params_with_values', 'intro_func_all_params_with_values_as_ordered_dict', 'code_params_with_values_to_signature_items_gen', 'code_params_with_values_to_signature']

from typing import Any, Callable, Awaitable, Dict, Generator, List, Optional, Tuple, NamedTuple, OrderedDict as OrderedDictType, Type, Union, Set, Sequence, cast
from types import ModuleType, CodeType, FrameType
import traceback
import inspect
from inspect import getattr_static as inspect__getattr_static
import sys
from cengal.code_flow_control.python_bytecode_manipulator import CodeParamNames, code_param_names, has_code, get_code, code_name
from cengal.text_processing.brackets_processing import Bracket, BracketPair, replace_text_with_brackets, find_text_in_brackets
from collections import OrderedDict
from importlib import import_module


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class WrongDepth(Exception):
    pass


class CanNotRetrieveFrame(Exception):
    pass


def frame(depth: Optional[int] = 1) -> FrameType:
    """
    :param depth: 0 - frame of this function, 1 - frame of the caller function, etc.
    :return:
    """
    depth = depth or 0
    if depth < 0:
        raise WrongDepth(depth)

    result = inspect.currentframe()
    if result is None:
        raise CanNotRetrieveFrame()

    for i in range(depth):
        result = result.f_back

    return result


def get_exception() -> Exception:
    ex_type, ex_value, ex_traceback = sys.exc_info()
    return ex_value.with_traceback(ex_traceback)


def get_exception_tripple() -> Tuple[Type, Exception, Any]:
    return sys.exc_info()


def exception_to_printable_text(exception: Exception) -> str:
    # return ''.join(traceback.format_exception(type(exception), exception, exception.__traceback__))
    return ''.join(traceback.TracebackException.from_exception(exception).format())


def is_async(entity) -> bool:
    return inspect.iscoroutine(entity) or inspect.isgenerator(entity) or inspect.iscoroutinefunction(entity) or inspect.isgeneratorfunction(entity) or inspect.isasyncgen(entity) or inspect.isasyncgenfunction(entity) or inspect.isawaitable(entity)


def is_callable(entity) -> bool:
    return callable(entity)


def func_param_names(func) -> CodeParamNames:
    return code_param_names(get_code(func))


def entity_arguments_description(entity: Callable) -> Tuple[Set[str], Sequence[str], Sequence[str], Sequence[str]]:
    init_code = get_code(entity)
    cpn: CodeParamNames = code_param_names(init_code)
    positional = cpn.positional
    positional_only = cpn.positional_only
    keyword_only = cpn.keyword_only
    all: Set[str] = set(positional) | set(positional_only) | set(keyword_only)
    return all, positional, positional_only, keyword_only


def func_code_name(func):
    code = get_code(func)
    return code_name(code)


def func_name(func):
    try:
        return func.__name__
    except AttributeError:
        return str()


def func_qualname(func):
    try:
        return func.__qualname__
    except AttributeError:
        return str()


def entity_name(entity):
    if isinstance(entity, FrameType):
        entity = entity.f_code
    
    if isinstance(entity, CodeType):
        return code_name(entity)
    else:
        return func_name(entity)


def entity_qualname(entity):
    if isinstance(entity, FrameType):
        entity = entity.f_code
    
    if isinstance(entity, CodeType):
        if sys.version_info >= (3, 11):
            return code_qualname(entity)
        else:
            # raise RuntimeError('CodeType.__qualname__ is available only since Python 3.11')
            entity_instance = find_entity(entity)
            return f'{entity_owner_name(entity_instance)}.{entity_name(entity_instance)}'
    else:
        return func_qualname(entity)


def entity_class(func) -> Optional[Type]:
    if inspect.ismethod(func):
        for func_class in inspect.getmro(func.__self__.__class__):
            if func_class.__dict__.get(func.__name__) is func:
                return func_class
        
        func_of_the_bound_method = func.__func__
        for func_class in inspect.getmro(func.__self__.__class__):
            if func_class.__dict__.get(func.__name__) is func_of_the_bound_method:
                return func_class
        
        func = func_of_the_bound_method
    
    if inspect.isfunction(func):
        try:
            func_class = getattr_ex(inspect.getmodule(func), func.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
        except AttributeError:
            return None
        
        if isinstance(func_class, type):
            return func_class
    
    return getattr_ex(func, '__objclass__', None)


def entity_owner(func) -> Optional[Union[Type, ModuleType]]:
    func_module = None
    if inspect.ismethod(func):
        for func_class in inspect.getmro(func.__self__.__class__):
            if func_class.__dict__.get(func.__name__) is func:
                return func_class
        
        func_of_the_bound_method = func.__func__
        for func_class in inspect.getmro(func.__self__.__class__):
            if func_class.__dict__.get(func.__name__) is func_of_the_bound_method:
                return func_class
        
        func = func_of_the_bound_method
        
    if inspect.isfunction(func):
        func_module = inspect.getmodule(func)
        try:
            func_class = getattr_ex(func_module, func.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
        except AttributeError:
            return func_module
        
        if isinstance(func_class, type):
            return func_class
        else:
            return func_module
    
    try:
        return getattr_ex(func, '__objclass__')
    except AttributeError:
        if func_module is None:
            func_module = inspect.getmodule(func)
        
        return func_module


module_repr_importable_str_bracket_pair: BracketPair = BracketPair([Bracket("<module '")], [Bracket("' from '")])
module_repr_full_file_path_bracket_pair: BracketPair = BracketPair([Bracket("' from '")], [Bracket("'>")])


def get_module_importable_str_and_path(module) -> Tuple[str, str]:
    if not inspect.ismodule(module):
        raise TypeError(f'Only modules are supported. {type(module)} was provided instead')
    
    module_repr: str = repr(module)
    importable_str = module_repr[find_text_in_brackets(module_repr, module_repr_importable_str_bracket_pair)]
    full_file_path = module_repr[find_text_in_brackets(module_repr, module_repr_full_file_path_bracket_pair)]
    return importable_str, full_file_path


def entity_owning_module_info_and_owning_path(entity) -> Tuple[ModuleType, str, str, List[Union[Type, ModuleType]]]:
    owning_path: List = list()
    module = None
    owner_is_module: bool = False
    while not owner_is_module:
        module = entity_owner(entity)
        owning_path.append(module)
        entity = module
        owner_is_module = inspect.ismodule(module)
    
    module_importable_str, module_file_full_path = get_module_importable_str_and_path(module)
    return module, module_importable_str, module_file_full_path, owning_path


def entity_owning_module_importable_str(entity) -> str:
    _, module_importable_str, _, _ = entity_owning_module_info_and_owning_path(entity)
    return module_importable_str


def entity_module_importable_str_and_owning_names_path(entity) -> Tuple[str, List[str]]:
    _, module_importable_str, _, owning_path = entity_owning_module_info_and_owning_path(entity)
    owning_path = owning_path[:-1]
    owning_names_path = [entity_name(owner) for owner in owning_path]
    return module_importable_str, owning_names_path


def entity_by_name_module_importable_str_and_owning_names_path(entity_name: str, module_importable_str: str, owning_names_path: List[str]) -> Any:
    owner = import_module(module_importable_str)
    for owner_name in reversed(owning_names_path):
        owner = getattr(owner, owner_name)

    return getattr(owner, entity_name)


module_repr_limited_bracket_pair: BracketPair = BracketPair([Bracket(" from '")], [Bracket("'>")])


def normalized_owner_repr(owner: Optional[Union[Type, ModuleType]]) -> str:
    if owner is None:
        return str()
    
    owner_repr: str = repr(owner)
    result, _ = replace_text_with_brackets(owner_repr, module_repr_limited_bracket_pair, ">", 1)
    return result


def normalized_code_owner_repr(code: CodeType, owner: Optional[Union[Type, ModuleType]]) -> str:
    if owner is None:
        return str()
    
    owner_repr: str = repr(owner)
    result, _ = replace_text_with_brackets(owner_repr, module_repr_limited_bracket_pair, f" line {code.co_firstlineno}>", 1)
    return result


def entity_owner_repr(entity):
    if isinstance(entity, FrameType):
        entity = entity.f_code
    
    if isinstance(entity, CodeType):
        return normalized_code_owner_repr(entity, entity_owner(entity))
    else:
        return normalized_owner_repr(entity_owner(entity))


def owner_name(owner: Optional[Union[Type, ModuleType]]) -> str:
    if owner is None:
        return str()
    
    return owner.__name__


def entity_owner_name(entity):
    return owner_name(entity_owner(entity))


def frame_param_names(frame_instance: FrameType) -> CodeParamNames:
    return code_param_names(frame_instance.f_code)


def intro_func_param_names(depth: Optional[int] = 1) -> CodeParamNames:
    return frame_param_names(frame(depth + 1))


CodeParamsWithValues = CodeParamNames
ParamWithValue = Tuple[str, Any]


def fill_code_params_with_values(code_params: CodeParamNames, args, kwargs) -> Tuple[CodeParamsWithValues, Tuple, Dict]:
    positional, positional_only, keyword_only = code_params
    positional_len = len(positional)
    positional_only_len = len(positional_only)
    keyword_only_len = len(keyword_only)
    positional_only_delimiter_place: Optional[int] = None
    if positional_only_len:
        positional_only_delimiter_place = positional_only_len
    
    positional_values = tuple(((arg, args[index]) for index, arg in enumerate(positional)))
    if positional_only_len:
        positional_only_values = tuple(((arg, args[index]) for index, arg in enumerate(positional_only)))
    else:
        positional_only_values = tuple()
    
    if keyword_only_len:
        keyword_only_values = tuple(((arg, kwargs[arg]) for arg in keyword_only))
    else:
        keyword_only_values = tuple()

    result_args: Tuple = args[positional_len:]
    result_kwargs: Dict = {key: value for key, value in kwargs.items() if key not in keyword_only}
    
    return CodeParamsWithValues(positional_values, positional_only_values, keyword_only_values), result_args, result_kwargs


def func_params_with_values(func: Union[Callable, Awaitable], args, kwargs) -> Tuple[CodeParamsWithValues, Tuple, Dict]:
    return fill_code_params_with_values(func_param_names(func), args, kwargs)


def intro_frame_params_with_values(frame_instance) -> CodeParamsWithValues:
    positional, positional_only, keyword_only = frame_param_names(frame_instance)
    fr_locals = frame_instance.f_locals
    return CodeParamsWithValues(tuple(((arg, fr_locals[arg]) for arg in positional)), \
        tuple(((arg, fr_locals[arg]) for arg in positional_only)), \
        tuple(((arg, fr_locals[arg]) for arg in keyword_only)))


def intro_func_params_with_values(depth: Optional[int] = 1) -> CodeParamsWithValues:
    return intro_frame_params_with_values(frame(depth + 1))


def intro_frame_all_params_with_values(frame_instance) -> Tuple[ParamWithValue]:
    positional, _, keyword_only = frame_param_names(frame_instance)
    all_params = positional + keyword_only
    fr_locals = frame_instance.f_locals
    return tuple(((arg, fr_locals[arg]) for arg in all_params))


def intro_func_all_params_with_values(depth: Optional[int] = 1) -> Tuple[ParamWithValue]:
    return intro_frame_all_params_with_values(frame(depth + 1))


def intro_frame_all_params_with_values_as_ordered_dict(frame_instance) -> OrderedDictType[str, Any]:
    return OrderedDict(intro_frame_all_params_with_values(frame_instance))


def intro_func_all_params_with_values_as_ordered_dict(depth: Optional[int] = 1) -> OrderedDictType[str, Any]:
    return intro_frame_all_params_with_values_as_ordered_dict(frame(depth + 1))


def code_params_with_values_to_signature_items_gen(code_params_with_values: CodeParamsWithValues, verbose: bool = False) -> Generator[str, None, None]:
    positional, positional_only, keyword_only = code_params_with_values
    positional_len = len(positional)
    positional_only_len = len(positional_only)
    keyword_only_len = len(keyword_only)
    positional_only_delimiter_place: Optional[int] = None
    if positional_only_len:
        positional_only_delimiter_place = positional_only_len
    
    for index, arg in enumerate(positional):
        arg_name, arg_value = arg
        if (positional_only_delimiter_place is not None) and (index == positional_only_delimiter_place):
            yield '/'
        
        if verbose:
            yield f'{arg_name}={arg_value}'
        else:
            yield str(arg_value)
    
    if keyword_only_len:
        yield '*'
        for arg in keyword_only:
            arg_name, arg_value = arg
            yield f'{arg_name}={arg_value}'


def code_params_to_signature_items_gen(code_params: CodeParamNames) -> Generator[str, None, None]:
    positional, positional_only, keyword_only = code_params
    positional_len = len(positional)
    positional_only_len = len(positional_only)
    keyword_only_len = len(keyword_only)
    positional_only_delimiter_place: Optional[int] = None
    if positional_only_len:
        positional_only_delimiter_place = positional_only_len
    
    for index, arg in enumerate(positional):
        if (positional_only_delimiter_place is not None) and (index == positional_only_delimiter_place):
            yield '/'
        
        # yield arg[0]
        yield arg
    
    if keyword_only_len:
        yield '*'
        for arg in keyword_only:
            # yield arg[0]
            yield arg


def code_params_with_values_to_signature(params_with_values: CodeParamsWithValues, verbose: bool = False, formatted: bool = False) -> str:
    if formatted:
        delimiter = ',\n\t'
    else:
        delimiter = ', '
    
    return delimiter.join(code_params_with_values_to_signature_items_gen(params_with_values, verbose))


def code_params_to_signature(param_names: CodeParamNames, formatted: bool = False) -> str:
    if formatted:
        delimiter = ',\n\t'
    else:
        delimiter = ', '
    
    return delimiter.join(code_params_to_signature_items_gen(param_names))


def entity_repr_limited(entity, formatted: bool = False):
    if isinstance(entity, FrameType):
        entity = entity.f_code
    
    if isinstance(entity, CodeType):
        code = entity
    else:
        code = get_code(entity)
    
    func_name = code_name(code)
    param_names = code_param_names(code)
    function_params_str = code_params_to_signature(param_names, formatted)
    return f'{func_name}({function_params_str})'


def entity_repr_limited_try_qualname(entity, formatted: bool = False):
    if has_code(entity):
        if isinstance(entity, FrameType):
            entity = entity.f_code
        
        if isinstance(entity, CodeType):
            code = entity
            func_name = code_name(code)
        else:
            code = get_code(entity)
            func_name = func_qualname(entity)
        
        param_names = code_param_names(code)
        function_params_str = code_params_to_signature(param_names, formatted)
        return f'{func_name}({function_params_str})'
    else:
        entity_type_name = type(entity).__name__
        entity_properties_with_values = entity_properties_values(entity)
        if formatted:
            delimiter = ',\n\t'
        else:
            delimiter = ', '
        
        entity_properties_str = delimiter.join([f'{key}={value}' for key, value in entity_properties_with_values.items()])
        return f'{entity_type_name}({entity_properties_str})'


def entity_repr_owner_based(entity, formatted: bool = False):
    if isinstance(entity, FrameType):
        entity = entity.f_code
    
    owner = entity_owner(entity)
    if isinstance(entity, CodeType):
        owner_repr = normalized_code_owner_repr(entity, owner)
    else:
        owner_repr = normalized_owner_repr(owner)

    if isinstance(owner, ModuleType):
        _entity_repr_limited = entity_repr_limited_try_qualname
    else:
        _entity_repr_limited = entity_repr_limited
    
    if owner_repr:
        return f'{owner_repr}.{_entity_repr_limited(entity, formatted)}'
    else:
        return _entity_repr_limited(entity, formatted)


def entity_repr(entity, formatted: bool = False):
    if isinstance(entity, FrameType):
        entity = entity.f_code
    
    module = inspect.getmodule(entity)
    if isinstance(entity, CodeType):
        owner_repr = normalized_code_owner_repr(entity, module)
    else:
        owner_repr = normalized_owner_repr(module)
    
    if owner_repr:
        return f'{owner_repr}.{entity_repr_limited_try_qualname(entity, formatted)}'
    else:
        return entity_repr_limited_try_qualname(entity, formatted)


def entity_properties(entity) -> Set[str]:
    """
    Example:
        from cengal.parallel_execution.coroutines.coro_standard_services.wait_coro import WaitCoroRequest
        from cengal.introspection.inspect import entity_properties
        
        print(entity_properties(WaitCoroRequest))
        >> {'__weakref__', 'put_list', 'fastest', 'atomic', 'list', 'single', 'put_fastest', '_save', 'put_single', 'put_atomic'}


        print(entity_properties(WaitCoroRequest()))
        >> {'result_required', 'args', 'tree', 'kill_on_timeout', 'timeout', 'request_type', 'kwargs', 'provide_to_request_handler'}


        def my_func(a, b, *, c, d):
            return a + b + c + d

        my_func.my_property = 2

        print(entity_properties(my_func))
        >> {'my_property'}


    Args:
        entity (_type_): _description_

    Returns:
        Set[str]: _description_
    """    
    entity_type_items = set(dir(type(entity)))
    entity_items = set(dir(entity))
    return entity_items - entity_type_items


def entity_properties_values(entity) -> Dict[str, Any]:
    result = dict()
    for property_name in sorted(entity_properties(entity)):
        result[property_name] = getattr(entity, property_name)
    
    return result


def class_properties(entity: Type) -> Set[str]:
    """
    Example:

    Args:
        entity (_type_): _description_

    Returns:
        Set[str]: _description_
    """    
    entity_type_items = set(dir(type(entity)))
    entity_items = set(dir(entity))
    mro = inspect.getmro(entity)
    base_members: Set = set()
    for base in mro:
        if base is entity:
            continue
        
        base_members |= set(dir(base))
    
    return (entity_items - entity_type_items) - base_members


def class_properties_values(entity) -> Dict[str, Any]:
    result = dict()
    for property_name in sorted(class_properties(entity)):
        result[property_name] = getattr(entity, property_name)
    
    return result


def class_properties_withot_object(entity: Type) -> Set[str]:
    """
    Example:

    Args:
        entity (_type_): _description_

    Returns:
        Set[str]: _description_
    """    
    object_type_items = set(dir(object))
    entity_items = set(dir(entity))
    return entity_items - object_type_items


def class_properties_withot_object_values(entity) -> Dict[str, Any]:
    result = dict()
    for property_name in sorted(class_properties_withot_object(entity)):
        result[property_name] = getattr(entity, property_name)
    
    return result


def class_properties_including_overrided(entity: Type) -> Set[str]:
    """
    Example:

    Args:
        entity (_type_): _description_

    Returns:
        Set[str]: _description_
    """    
    entity_type_items = set(dir(type(entity)))
    entity_items = set(entity.__dict__.keys())
    return entity_items - entity_type_items


def class_properties_values_including_overrided(entity) -> Dict[str, Any]:
    result = dict()
    for property_name in sorted(class_properties_including_overrided(entity)):
        result[property_name] = getattr(entity, property_name)
    
    return result


def intro_frame_repr_limited(frame_instance, verbose: bool = False):
    func_name = code_name(frame_instance.f_code)
    params_with_values = intro_frame_params_with_values(frame_instance)
    function_params_str = code_params_with_values_to_signature(params_with_values, verbose)
    return f'{func_name}({function_params_str})'


def intro_frame_repr(frame_instance, verbose: bool = False):
    code = frame_instance.f_code
    func_name = code_name(code)
    module = inspect.getmodule(code)
    owner_repr = normalized_code_owner_repr(code, module)
    params_with_values = intro_frame_params_with_values(frame_instance)
    function_params_str = code_params_with_values_to_signature(params_with_values, verbose)
    return f'{owner_repr}.{func_name}({function_params_str})'


def intro_func_repr_limited(verbose: bool = False, depth: Optional[int] = 1):
    return intro_frame_repr_limited(frame(depth + 1), verbose)


def print_intro_func_repr_limited(verbose: bool = False, depth: Optional[int] = 1):
    print(intro_func_repr_limited(verbose, depth + 1))


pifrl = print_intro_func_repr_limited


def intro_func_repr(verbose: bool = False, depth: Optional[int] = 1):
    return intro_frame_repr(frame(depth + 1), verbose)


def print_intro_func_repr(verbose: bool = False, depth: Optional[int] = 1):
    print(intro_func_repr(verbose, depth + 1))


pifr = print_intro_func_repr


def get_str_of_data_info(data):
    """Returns string with data info: type and value

    Args:
        data (_type_): _description_

    Returns:
        _type_: _description_
    """    
    return f'type: {type(data)}; value: {data}'


gsodi = get_str_of_data_info


def get_multistr_of_data_info(data, shift_num: Optional[int] = None, shift_char: str = '\t'):
    """Returns multiline string with data info: type and value

    Args:
        data (_type_): _description_

    Returns:
        _type_: _description_
    """
    shift_num = shift_num or 0
    shift_str = shift_char * shift_num
    from pprint import pformat
    data_str = pformat(data)
    data_str_lines = data_str.split('\n')
    data_str = '\n'.join([f'{shift_str}{line}' for line in data_str_lines])
    return f'type: {type(data)}; value: \n{data_str}'


gmsodi = get_multistr_of_data_info


def get_multistr_of_data_value(data, shift_num: Optional[int] = None, shift_char: str = '\t'):
    """Returns multiline string with data value

    Args:
        data (_type_): _description_

    Returns:
        _type_: _description_
    """
    shift_num = shift_num or 0
    shift_str = shift_char * shift_num
    from pprint import pformat
    data_str = pformat(data)
    data_str_lines = data_str.split('\n')
    data_str = '\n'.join([f'{shift_str}{line}' for line in data_str_lines])
    return data_str


gmsodv = get_multistr_of_data_value


def print_data_info(data):
    """Print data info: type and value

    Args:
        data (_type_): _description_
    """    
    print(get_str_of_data_info(data))


pdi = print_data_info


def get_str_of_data_info_named(name, data):
    return f'<<{name}>> type: {type(data)}; value: {data}'


gsodin = get_str_of_data_info_named


def print_data_info_named(name, data):
    print(get_str_of_data_info_named(name, data))


pdin = print_data_info_named


def get_str_of_data_info_by_name(name, depth: Optional[int] = 1):
    fr = frame(depth + 1)
    data = fr.f_locals[name]
    return get_str_of_data_info_named(name, data)


gsodibn = get_str_of_data_info_by_name


def print_data_info_by_name(name, depth: Optional[int] = 1):
    print(get_str_of_data_info_by_name(name, depth + 1))


pdibn = print_data_info_by_name


def is_descriptor(entity):
    return hasattr(entity, '__get__') or hasattr(entity, '__set__') or hasattr(entity, '__delete__')


def is_setable_data_descriptor(entity):
    return hasattr(entity, '__get__') and hasattr(entity, '__set__')


def is_filled_descriptor(owning_object, entity) -> bool:
    """
        class _foo:
        __slots__ = ['foo', 'bar']

        def __init__(self):
            self.bar = 2

    'foo' - not filled
    'bar' - filled

    Args:
        owning_object (_type_): _description_
        entity (_type_): _description_

    Returns:
        _type_: _description_
    """    
    try:
        entity.__get__(owning_object)
    except AttributeError:
        return False
    
    return True


def filled_slots_names(owning_object) -> Optional[List[str]]:
    try:
        slots_names = owning_object.__slots__
    except AttributeError:
        return None
    
    result = list()
    for slot_name in slots_names:
        slot = inspect.getattr_static(owning_object, slot_name)
        if is_filled_descriptor(owning_object, slot):
            result.append(slot_name)
    
    return result


def filled_slots(owning_object) -> Optional[List[Any]]:
    try:
        slots_names = owning_object.__slots__
    except AttributeError:
        return None
    
    result = list()
    for slot_name in slots_names:
        slot = inspect.getattr_static(owning_object, slot_name)
        if is_filled_descriptor(owning_object, slot):
            result.append(slot)
    
    return result


def filled_slots_with_names(owning_object) -> Optional[List[Tuple[str, Any]]]:
    try:
        slots_names = owning_object.__slots__
    except AttributeError:
        return None
    
    result = list()
    for slot_name in slots_names:
        slot = inspect.getattr_static(owning_object, slot_name)
        if is_filled_descriptor(owning_object, slot):
            result.append((slot_name, slot))
    
    return result


def filled_slots_with_names_gen(owning_object) -> Generator[Tuple[str, Any], Any, Any]:
    try:
        slots_names = owning_object.__slots__
        for slot_name in slots_names:
            slot = inspect__getattr_static(owning_object, slot_name)
            try:
                slot.__get__(owning_object)
                yield (slot_name, slot)
            except AttributeError:
                pass
    except AttributeError:
        pass


def filled_slot_names_with_values_gen(owning_object) -> Generator[Tuple[str, Any], Any, Any]:
    try:
        slots_names = owning_object.__slots__
        for slot_name in slots_names:
            try:
                yield (slot_name, getattr(owning_object, slot_name))
            except AttributeError:
                pass
    except AttributeError:
        pass


def current_entity_name(depth: Optional[int] = 1):
    fr = frame(depth + 1)
    return entity_name(fr)
    # entity_name = code_name(fr.f_code)
    # # print(dir(fr.f_code))
    # # print(fr.f_code.__class__)
    # # print(dir(fr))
    # # print(fr.__class__)
    # return entity_name


cen = current_entity_name


def current_entity_owner_name(depth: Optional[int] = 1):
    fr = frame(depth + 1)
    return owner_name(entity_owner(fr.f_code))


# def current_entity_full_name(depth: Optional[int] = 1):
#     fr = frame(depth + 1)
#     entity_name = code_name(fr.f_code)
#     return entity_name


# cefn = current_entity_full_name


def getattr_ex(entity, attr_name: str) -> Any:
    try:
        return getattr(entity, attr_name)
    except AttributeError:  # See code of the `inspect.getmembers()`
        if attr_name in entity.__dict__:
            return entity.__dict__[attr_name]
        else:
            raise AttributeError(f'Attribute "{attr_name}" was not found in the entity "{entity}"')


def entity_is_function(entity) -> bool:
    owner: Optional[Union[Type, ModuleType]] = entity_owner(entity)
    if owner is None:
        return False
    
    if inspect.ismodule(owner):
        entity_name_str: str = entity_name(entity)
        if hasattr(owner, entity_name_str):
            possible_entity = getattr_ex(owner, entity_name_str)
            possible_entity_code = get_code(possible_entity)
            entity_code = get_code(entity)
            if possible_entity_code is entity_code:
                return True
            else:
                return False
        else:
            return False
    else:
        return False


def get_function_by_entity(entity) -> Optional[Callable]:
    owner: Optional[Union[Type, ModuleType]] = entity_owner(entity)
    if owner is None:
        return None
    
    if inspect.ismodule(owner):
        entity_name_str: str = entity_name(entity)
        if hasattr(owner, entity_name_str):
            possible_entity = getattr_ex(owner, entity_name_str)
            possible_entity_code = get_code(possible_entity)
            entity_code = get_code(entity)
            if possible_entity_code is entity_code:
                return possible_entity
            else:
                return None
        else:
            return None
    else:
        return None


def entity_is_method(entity) -> bool:
    owner: Optional[Union[Type, ModuleType]] = entity_owner(entity)
    if owner is None:
        return False
    
    if inspect.isclass(owner):
        entity_name_str: str = entity_name(entity)
        if hasattr(owner, entity_name_str):
            possible_entity = getattr_ex(owner, entity_name_str)
            possible_entity_code = get_code(possible_entity)
            entity_code = get_code(entity)
            if possible_entity_code is entity_code:
                return True
            else:
                return False
        else:
            return False
    else:
        return False


def get_method_by_entity(entity) -> Optional[Callable]:
    owner: Optional[Union[Type, ModuleType]] = entity_owner(entity)
    if owner is None:
        return None
    
    if inspect.isclass(owner):
        entity_name_str: str = entity_name(entity)
        if hasattr(owner, entity_name_str):
            possible_entity = getattr_ex(owner, entity_name_str)
            possible_entity_code = get_code(possible_entity)
            entity_code = get_code(entity)
            if possible_entity_code is entity_code:
                return possible_entity
            else:
                return None
        else:
            return None
    else:
        return None


def entity_is_unbound_method(entity):
    owner: Optional[Union[Type, ModuleType]] = entity_owner(entity)
    if owner is None:
        return False
    
    if inspect.ismodule(owner):
        entity_name_str: str = entity_name(entity)
        if hasattr(owner, entity_name_str):
            possible_entity = getattr_ex(owner, entity_name_str)
            possible_entity_code = get_code(possible_entity)
            entity_code = get_code(entity)
            if possible_entity_code is entity_code:
                return False
            else:
                return True
        else:
            return True
    else:
        return False


def get_unbound_method_by_entity(entity) -> Optional[Callable]:
    raise NotImplementedError()


class AnAppropriateOwnerParameterWasNotFoundError(Exception):
    pass


class PossibleOwnerParameterDoesNotMatchError(AnAppropriateOwnerParameterWasNotFoundError):
    pass


class EntityHasNoPositionalParametersError(AnAppropriateOwnerParameterWasNotFoundError):
    pass


def get_owner_parameter(entity, owner_parameter_name: str, return_even_if_not_match: bool = False):
    if entity_is_function(entity):
        raise TypeError(f'Function has no {owner_parameter_name} parameter')
    
    if entity_is_method(entity):
        return entity.__self__
    elif entity_is_unbound_method(entity):
        if isinstance(entity, FrameType):
            entity_name_str: str = entity_name(entity)
            params_with_values: CodeParamsWithValues = intro_frame_params_with_values(entity)
            positional, positional_only, keyword_only = params_with_values
            if positional:
                possible_positional_self = positional[0][1]
            elif positional_only:
                possible_positional_self = positional_only[0][1]
            else:
                raise EntityHasNoPositionalParametersError
            
            if hasattr(possible_positional_self, entity_name_str):
                possible_entity = getattr_ex(possible_positional_self, entity_name_str)
                possible_entity_code = get_code(possible_entity)
                entity_code = get_code(entity)
                if possible_entity_code is entity_code:
                    return possible_positional_self
                else:
                    if return_even_if_not_match:
                        return possible_positional_self
                    else:
                        raise PossibleOwnerParameterDoesNotMatchError
            
            raise AnAppropriateOwnerParameterWasNotFoundError(f'Can not find an appropriate {owner_parameter_name} parameter in {entity}')
    else:
        raise TypeError(f'{entity} of type {type(entity)} is not supported')


def get_self_parameter(entity):
    return get_owner_parameter(entity, 'self')


def get_cls_parameter(entity):
    return get_owner_parameter(entity, 'cls')


def get_any_owner_parameter(entity, owner_parameter_name: str, any_positional: bool = True, any_keyword: bool = True):
    if entity_is_function(entity):
        raise TypeError(f'Function has no {owner_parameter_name} parameter')
    
    if entity_is_method(entity):
        return entity.__self__
    elif entity_is_unbound_method(entity):
        if isinstance(entity, FrameType):
            entity_name_str: str = entity_name(entity)
            params_with_values: CodeParamsWithValues = intro_frame_params_with_values(entity)
            positional, positional_only, keyword_only = params_with_values
            
            if any_positional:
                all_positional = positional + positional_only
                for param in all_positional:
                    possible_positional_self_name, possible_positional_self = param
                    if hasattr(possible_positional_self, entity_name_str):
                        possible_entity = getattr_ex(possible_positional_self, entity_name_str)
                        possible_entity_code = get_code(possible_entity)
                        entity_code = get_code(entity)
                        if possible_entity_code is entity_code:
                            return possible_positional_self
            else:
                if positional:
                    possible_positional_self = positional[0][1]
                elif positional_only:
                    possible_positional_self = positional_only[0][1]
                
                if hasattr(possible_positional_self, entity_name_str):
                    possible_entity = getattr_ex(possible_positional_self, entity_name_str)
                    possible_entity_code = get_code(possible_entity)
                    entity_code = get_code(entity)
                    if possible_entity_code is entity_code:
                        return possible_positional_self
            
            if any_keyword:
                for possible_keyword_self_name, possible_keyword_self in keyword_only:
                    if hasattr(possible_keyword_self, entity_name_str):
                        possible_entity = getattr_ex(possible_keyword_self, entity_name_str)
                        possible_entity_code = get_code(possible_entity)
                        entity_code = get_code(entity)
                        if possible_entity_code is entity_code:
                            return possible_keyword_self
            else:
                keyword_only_dict: Dict[str, Any] = dict(keyword_only)
                if owner_parameter_name in keyword_only_dict:
                    possible_keyword_self = keyword_only_dict[owner_parameter_name]
                    if hasattr(possible_keyword_self, entity_name_str):
                        possible_entity = getattr_ex(possible_keyword_self, entity_name_str)
                        possible_entity_code = get_code(possible_entity)
                        entity_code = get_code(entity)
                        if possible_entity_code is entity_code:
                            return possible_keyword_self
            
            raise AnAppropriateOwnerParameterWasNotFoundError(f'Can not find {owner_parameter_name} parameter in {entity}')


def get_any_self_parameter(entity, any_positional: bool = True, any_keyword: bool = True):
    return get_owner_parameter(entity, 'self', any_positional, any_keyword)


def get_any_cls_parameter(entity, any_positional: bool = True, any_keyword: bool = True):
    return get_owner_parameter(entity, 'cls' , any_positional, any_keyword)


def find_method_in_module_by_code(module, code_to_find: CodeType) -> Optional[CodeType]:
    for entity_name_str in dir(module):
        entity = getattr_ex(module, entity_name_str)
        if inspect.isclass(entity):
            possible_method = find_method_in_class_by_code(entity, code_to_find)
            if possible_method is not None:
                return possible_method
    
    return None


def find_method_in_class_by_code(class_to_search_in, code_to_find: CodeType) -> Optional[CodeType]:
    subclassess = list()
    for entity_name_str in class_properties_including_overrided(class_to_search_in):
        entity = getattr_ex(class_to_search_in, entity_name_str)
        if inspect.isclass(entity):
            subclassess.append(entity)
        elif inspect.isfunction(entity) or inspect.ismethod(entity):
            if get_code(entity) is code_to_find:
                return entity
        
    for subclass in subclassess:
        possible_method = find_method_in_class_by_code(subclass, code_to_find)
        if possible_method is not None:
            return possible_method
    
    return None


class EntityWasNotFoundError(Exception):
    pass


def find_entity(entity: Union[Callable, FrameType, CodeType]) -> Callable:
    if not (inspect.isfunction(entity) or inspect.ismethod(entity) or isinstance(entity, FrameType) or isinstance(entity, CodeType)):
        raise TypeError(f'Only functions, methods, frames and codes are supported. {type(entity)} was provided instead')
    
    result = get_function_by_entity(entity)
    if result is not None:
        return result
    
    result = get_method_by_entity(entity)
    if result is not None:
        return result
    
    entity_name_str: str = entity_name(entity)
    entity_instance: Callable = None
    if isinstance(entity, FrameType):
        owner = None
        need_to_try_clr: bool = False
        try:
            owner = get_self_parameter(entity)
        except AnAppropriateOwnerParameterWasNotFoundError:
            need_to_try_clr = True
        
        if need_to_try_clr:
            try:
                owner = get_cls_parameter(entity)
            except AnAppropriateOwnerParameterWasNotFoundError:
                pass
        
        entity_code = entity.f_code
        if owner is not None:
            entity_instance = getattr_ex(owner, entity_name_str)
            if get_code(entity_instance) is entity_code:
                return entity_instance
        
        entity = entity_code
    
    if isinstance(entity, CodeType):
        entity_owner_instance = entity_owner(entity)
        entity_instance = find_method_in_module_by_code(entity_owner_instance, entity)
        if entity_instance is not None:
            return entity_instance
    
    raise EntityWasNotFoundError


def find_current_entity(depth: Optional[int] = 1) -> Callable:
    return find_entity(frame(depth + 1))



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['RelativePath', 'relative_to_src', 'path_relative_to_src', 'relative_to_cwd', 
           'path_relative_to_cwd', 'WrongBaseDirError', 'get_relative_path_part', 'sep', 
           'canonical_path']

import os
from os.path import normpath, expanduser, realpath, abspath, normcase, sep
from typing import Optional
from cengal.file_system.directory_manager import current_src_dir
from cengal.text_processing.text_processing import removeprefix

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class RelativePath:
    def __init__(self, base_path: str):
        self.base_path: str = base_path

    def __call__(self, relative_path: str):
        return normpath(os.path.join(self.base_path, relative_path))


def relative_to_src(depth: Optional[int] = 1) -> RelativePath:
    """

    :param depth: 0 - path of this file, 1 - path of the caller's file, etc.
    :return:
    """
    depth = depth or 0
    depth += 1
    return RelativePath(current_src_dir(depth))


def path_relative_to_src(relative_path: str, depth: Optional[int] = 1) -> str:
    """

    :param relative_path:
    :param depth: 0 - path of this file, 1 - path of the caller's file, etc.
    :return:
    """
    depth = depth or 0
    depth += 1
    return RelativePath(current_src_dir(depth))(relative_path)


def relative_to_cwd() -> RelativePath:
    return RelativePath(os.getcwd())


def path_relative_to_cwd(relative_path: str) -> str:
    return RelativePath(os.getcwd())(relative_path)


class WrongBaseDirError(Exception):
    pass


def get_relative_path_part(path: str, base_dir: str) -> str:
    path = normpath(path)
    base_dir = normpath(base_dir)
    if not path.startswith(base_dir):
        raise WrongBaseDirError
    
    relative_part = removeprefix(path, base_dir)
    if relative_part.startswith(sep):
        relative_part = removeprefix(relative_part, sep)
    
    return relative_part


def canonical_path(path: str, resolve_symlinks: bool = True) -> str:
    """
    Convert a path to its canonical, case-normalized, absolute version.
    """

    path = expanduser(path)
    if resolve_symlinks:
        path = realpath(path)
    else:
        path = abspath(path)
    
    return normcase(path)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


cdef extern from "memory_access.h":
    void c_write_uint64(unsigned long long base_address, unsigned long long offset, unsigned long long value) nogil
    unsigned long long c_read_uint64(unsigned long long base_address, unsigned long long offset) nogil
    void c_write_int64(unsigned long long base_address, unsigned long long offset, long long value) nogil
    long long c_read_int64(unsigned long long base_address, unsigned long long offset) nogil
    void c_write_uint32(unsigned long long base_address, unsigned long long offset, unsigned int value) nogil
    unsigned int c_read_uint32(unsigned long long base_address, unsigned long long offset) nogil
    void c_write_int32(unsigned long long base_address, unsigned long long offset, int value) nogil
    int c_read_int32(unsigned long long base_address, unsigned long long offset) nogil
    void c_write_uint16(unsigned long long base_address, unsigned long long offset, unsigned short value) nogil
    unsigned short c_read_uint16(unsigned long long base_address, unsigned long long offset) nogil
    void c_write_int16(unsigned long long base_address, unsigned long long offset, short value) nogil
    short c_read_int16(unsigned long long base_address, unsigned long long offset) nogil
    void c_write_uint8(unsigned long long base_address, unsigned long long offset, unsigned char value) nogil
    unsigned char c_read_uint8(unsigned long long base_address, unsigned long long offset) nogil
    void c_write_int8(unsigned long long base_address, unsigned long long offset, char value) nogil
    char c_read_int8(unsigned long long base_address, unsigned long long offset) nogil
    void c_write_float(unsigned long long base_address, unsigned long long offset, float value) nogil
    float c_read_float(unsigned long long base_address, unsigned long long offset) nogil
    void c_write_double(unsigned long long base_address, unsigned long long offset, double value) nogil
    double c_read_double(unsigned long long base_address, unsigned long long offset) nogil
    void c_copy_memory(unsigned long long base_address, unsigned long long offset, unsigned long long size, unsigned long long source_offset) nogil
    void c_copy_memory_from(unsigned long long base_address, unsigned long long offset, unsigned long long size, unsigned long long source_offset) nogil
    void c_set_memory(unsigned long long base_address, unsigned long long offset, unsigned long long size, unsigned char value) nogil
    void c_zero_memory(unsigned long long base_address, unsigned long long offset, unsigned long long size) nogil
    unsigned long long c_mask_least_significant_bits(long long value, int n) nogil

cpdef void write_uint64(unsigned long long base_address, unsigned long long offset, unsigned long long value) noexcept:
    c_write_uint64(base_address, offset, value)

cpdef unsigned long long read_uint64(unsigned long long base_address, unsigned long long offset) noexcept:
    return c_read_uint64(base_address, offset)

cpdef void write_int64(unsigned long long base_address, unsigned long long offset, long long value) noexcept:
    c_write_int64(base_address, offset, value)

cpdef long long read_int64(unsigned long long base_address, unsigned long long offset) noexcept:
    return c_read_int64(base_address, offset)

cpdef void write_uint32(unsigned long long base_address, unsigned long long offset, unsigned int value) noexcept:
    c_write_uint32(base_address, offset, value)

cpdef unsigned int read_uint32(unsigned long long base_address, unsigned long long offset) noexcept:
    return c_read_uint32(base_address, offset)

cpdef void write_int32(unsigned long long base_address, unsigned long long offset, int value) noexcept:
    c_write_int32(base_address, offset, value)

cpdef int read_int32(unsigned long long base_address, unsigned long long offset) noexcept:
    return c_read_int32(base_address, offset)

cpdef void write_uint16(unsigned long long base_address, unsigned long long offset, unsigned short value) noexcept:
    c_write_uint16(base_address, offset, value)

cpdef unsigned short read_uint16(unsigned long long base_address, unsigned long long offset) noexcept:
    return c_read_uint16(base_address, offset)

cpdef void write_int16(unsigned long long base_address, unsigned long long offset, short value) noexcept:
    c_write_int16(base_address, offset, value)

cpdef short read_int16(unsigned long long base_address, unsigned long long offset) noexcept:
    return c_read_int16(base_address, offset)

cpdef void write_uint8(unsigned long long base_address, unsigned long long offset, unsigned char value) noexcept:
    c_write_uint8(base_address, offset, value)

cpdef unsigned char read_uint8(unsigned long long base_address, unsigned long long offset) noexcept:
    return c_read_uint8(base_address, offset)

cpdef void write_int8(unsigned long long base_address, unsigned long long offset, char value) noexcept:
    c_write_int8(base_address, offset, value)

cpdef char read_int8(unsigned long long base_address, unsigned long long offset) noexcept:
    return c_read_int8(base_address, offset)

cpdef void write_float(unsigned long long base_address, unsigned long long offset, float value) noexcept:
    c_write_float(base_address, offset, value)

cpdef float read_float(unsigned long long base_address, unsigned long long offset) noexcept:
    return c_read_float(base_address, offset)

cpdef void write_double(unsigned long long base_address, unsigned long long offset, double value) noexcept:
    c_write_double(base_address, offset, value)

cpdef double read_double(unsigned long long base_address, unsigned long long offset) noexcept:
    return c_read_double(base_address, offset)

cpdef void copy_memory(unsigned long long base_address, unsigned long long offset, unsigned long long size, unsigned long long source_offset) noexcept:
    c_copy_memory(base_address, offset, size, source_offset)

cpdef void copy_memory_from(unsigned long long base_address, unsigned long long offset, unsigned long long size, unsigned long long source_offset) noexcept:
    c_copy_memory_from(base_address, offset, size, source_offset)

cpdef void set_memory(unsigned long long base_address, unsigned long long offset, unsigned long long size, unsigned char value) noexcept:
    c_set_memory(base_address, offset, size, value)

cpdef void zero_memory(unsigned long long base_address, unsigned long long offset, unsigned long long size) noexcept:
    c_zero_memory(base_address, offset, size)

cpdef unsigned long long mask_least_significant_bits(long long value, int n) noexcept:
    return c_mask_least_significant_bits(value, n)

# cpdef unsigned long long list__get_item__impl(unsigned long long key, unsigned long long base_address, unsigned long long offset__pointer_to_internal_list, object get_obj_callback, unsigned long long & item_type, unsigned long long & need_to_raise_exception):
#     cdef unsigned long long pointer_to_internal_list
#     cdef unsigned long long item_type_offset
#     cdef unsigned long long item_value_offset
#     pointer_to_internal_list = c_read_uint64(base_address, offset__pointer_to_internal_list)
#     item_type_offset = pointer_to_internal_list + 32 + key * 16
#     item_value_offset = pointer_to_internal_list + 40 + key * 16
#     item_type = c_read_uint64(base_address, item_type_offset)
#     if item_type == 1:
#         return c_read_int64(base_address, item_value_offset)
#     elif item_type == 2:
#         return read_double(base_address, item_value_offset)
#     elif item_type == 3:
#         return c_read_uint64(base_address, item_value_offset)
#     elif item_type == 0:
#         item_offset = c_read_uint64(base_address, item_value_offset)
#         return get_obj_callback(item_offset)
#     else:
#         need_to_raise_exception = 1
#         return 0


cpdef object list__get_item(long long key, unsigned long long base_address, unsigned long long offset__pointer_to_internal_list, object get_obj_callback):
    cdef unsigned long long pointer_to_internal_list
    cdef unsigned long long item_type_offset
    cdef unsigned long long item_value_offset
    cdef unsigned long long item_type
    cdef unsigned long long unsigned_key
    cdef unsigned long long self_len
    cdef unsigned long long item_offset

    pointer_to_internal_list = c_read_uint64(base_address, offset__pointer_to_internal_list)
    self_len = read_uint64(base_address, pointer_to_internal_list + 24)
    if key < 0:
        key += self_len
    
    if key < 0 or key >= self_len:
        raise IndexError
    
    unsigned_key = key
    item_type_offset = pointer_to_internal_list + 32 + unsigned_key * 16
    item_value_offset = pointer_to_internal_list + 40 + unsigned_key * 16
    item_type = c_read_uint64(base_address, item_type_offset)
    
    if item_type == 2:
        return c_read_int64(base_address, item_value_offset)
    elif item_type == 3:
        return c_read_double(base_address, item_value_offset)
    elif item_type == 4:
        return bool(c_read_uint64(base_address, item_value_offset))
    elif item_type == 0:
        return None
    elif item_type == 1:
        item_offset = c_read_uint64(base_address, item_value_offset)
        return get_obj_callback(item_offset)
    else:
        raise ValueError


cpdef object list__get_item_as_offset(long long key, unsigned long long base_address, unsigned long long offset__pointer_to_internal_list):
    cdef unsigned long long pointer_to_internal_list
    cdef unsigned long long item_type_offset
    cdef unsigned long long item_value_offset
    cdef unsigned long long item_type
    cdef unsigned long long unsigned_key
    cdef unsigned long long self_len

    pointer_to_internal_list = c_read_uint64(base_address, offset__pointer_to_internal_list)
    self_len = read_uint64(base_address, pointer_to_internal_list + 24)
    if key < 0:
        key += self_len
    
    if key < 0 or key >= self_len:
        raise IndexError
    
    unsigned_key = key
    item_type_offset = pointer_to_internal_list + 32 + unsigned_key * 16
    item_value_offset = pointer_to_internal_list + 40 + unsigned_key * 16
    item_type = c_read_uint64(base_address, item_type_offset)
    
    if item_type == 2:
        return (item_type, c_read_uint64(base_address, item_value_offset))
    elif item_type == 3:
        return (item_type, c_read_uint64(base_address, item_value_offset))
    elif item_type == 4:
        return (item_type, c_read_uint64(base_address, item_value_offset))
    elif item_type == 0:
        return (item_type, 0)
    elif item_type == 1:
        return (item_type, c_read_uint64(base_address, item_value_offset))
    else:
        raise ValueError


cpdef void list__free_item(unsigned long long key, unsigned long long base_address, unsigned long long pointer_to_internal_list, object destroy_obj_method):
    cdef unsigned long long item_type_offset
    cdef unsigned long long item_type
    cdef unsigned long long unsigned_key
    cdef unsigned long long item_offset

    unsigned_key = key
    item_type = c_read_uint64(base_address, pointer_to_internal_list + 32 + unsigned_key * 16 + 0)
    if item_type == 1:
        item_offset = c_read_uint64(base_address, pointer_to_internal_list + 32 + unsigned_key * 16 + 8)
        destroy_obj_method(item_offset)
    elif item_type == 2:
        pass
    elif item_type == 3:
        pass
    elif item_type == 4:
        pass
    elif item_type == 0:
        pass
    else:
        raise ValueError

    item_type_offset = pointer_to_internal_list + 32 + unsigned_key * 16 + 0
    c_write_uint64(base_address, item_type_offset, 0)


cpdef void list__set_item(unsigned long long key, object value, unsigned long long base_address, unsigned long long offset__pointer_to_internal_list, unsigned int need_to_free_item, object destroy_obj_method, object put_obj_callback):
    cdef unsigned long long pointer_to_internal_list
    cdef unsigned long long item_type_offset
    cdef unsigned long long item_value_offset
    cdef unsigned long long item_type
    cdef unsigned long long unsigned_key
    cdef unsigned long long self_len
    cdef unsigned long long item_offset

    pointer_to_internal_list = read_uint64(base_address, offset__pointer_to_internal_list)
    self_len = read_uint64(base_address, pointer_to_internal_list + 24)
    if key < 0:
        key += self_len
    
    if key < 0 or key >= self_len:
        raise IndexError
    
    if need_to_free_item:
        list__free_item(key, base_address, pointer_to_internal_list, destroy_obj_method)
    
    unsigned_key = key
    item_type_offset = pointer_to_internal_list + 32 + unsigned_key * 16 + 0
    item_value_offset = pointer_to_internal_list + 32 + unsigned_key * 16 + 8
    if isinstance(value, int):
        c_write_int64(base_address, item_value_offset, value)
        item_type = 2
    elif isinstance(value, float):
        c_write_double(base_address, item_value_offset, value)
        item_type = 3
    elif isinstance(value, bool):
        c_write_uint64(base_address, item_value_offset, int(value))
        item_type = 4
    elif value is None:
        item_type = 0
    else:
        item_mapped_obj, item_offset, item_size = put_obj_callback(value)
        c_write_uint64(base_address, item_value_offset, item_offset)
        item_type = 1

    c_write_uint64(base_address, item_type_offset, item_type)


cpdef void list__set_item_as_offset(unsigned long long key, unsigned long long value_item_type, unsigned long long value_item_offset, unsigned long long base_address, unsigned long long offset__pointer_to_internal_list, unsigned int need_to_free_item, object destroy_obj_method):
    cdef unsigned long long pointer_to_internal_list
    cdef unsigned long long item_type_offset
    cdef unsigned long long item_value_offset
    cdef unsigned long long unsigned_key
    cdef unsigned long long self_len

    pointer_to_internal_list = read_uint64(base_address, offset__pointer_to_internal_list)
    self_len = read_uint64(base_address, pointer_to_internal_list + 24)
    if key < 0:
        key += self_len
    
    if key < 0 or key >= self_len:
        raise IndexError
    
    if need_to_free_item:
        list__free_item(key, base_address, pointer_to_internal_list, destroy_obj_method)
    
    unsigned_key = key
    item_type_offset = pointer_to_internal_list + 32 + unsigned_key * 16 + 0
    item_value_offset = pointer_to_internal_list + 32 + unsigned_key * 16 + 8
    if value_item_type == 2:
        c_write_int64(base_address, item_value_offset, value_item_offset)
    elif value_item_type == 3:
        c_write_double(base_address, item_value_offset, value_item_offset)
    elif value_item_type == 4:
        c_write_uint64(base_address, item_value_offset, value_item_offset)
    elif value_item_type == 0:
        pass
    else:
        c_write_uint64(base_address, item_value_offset, value_item_offset)
    
    c_write_uint64(base_address, item_type_offset, value_item_type)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'AsyncEventBus', 'AsyncEventBusRequest', 'EventID', 'Handler', 'try_send_async_event'
]


import sys
from enum import Enum
from typing import Dict, Set, List, Tuple, Union, Type, Optional, Any, Deque
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services_internal_lib.service_with_a_direct_request import *
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import *
from cengal.parallel_execution.coroutines.coro_standard_services.event_bus import EventID, Handler
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import *
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.introspection.inspect import get_exception
from cengal.time_management.repeat_for_a_time import Tracer
from collections import deque


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class AsyncEventBusRequest(ServiceRequest):
    def register_handler(self, event: EventID, handler: Handler) -> None:
        return self._save(0, event, handler)

    def remove_handler(self, event: EventID, handler: Handler) -> None:
        return self._save(1, event, handler)

    def send_event(self, event: EventID, data: Any,
                   priority: Optional[CoroPriority]=CoroPriority.low) -> None:
        return self._save(2, event, data, priority)

    def get_event(self, event: EventID, default: Any = None) -> ValueExistence[Any]:
        return self._save(3, event, default)

    def flush_event_data(self, event: EventID) -> ValueExistence[Any]:
        return self._save(5, event)

    def wait(self, event: EventID) -> Any:
        """
        Will block coroutine untill result is ready
        :param event:
        :return: event data
        """
        return self._save(4, event)


class AsyncEventBus(ServiceWithADirectRequestMixin, Service, EntityStatsMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(AsyncEventBus, self).__init__(loop)

        self._request_workers = {
            0: self._on_register_handler,
            1: self._on_remove_handler,
            2: self._on_send_event,
            3: self._on_get_event,
            4: self._on_wait,
            5: self._on_flush_event_data,
        }

        self.waiters: Dict[EventID, Set[CoroID]] = dict()
        self.events_by_waiter: Dict[CoroID, Set[Any]] = dict()
        self.handlers: Dict[EventID, Set[Handler]] = dict()
        self.events: Dict[CoroPriority, Dict[EventID, Deque[Any]]] = {
            CoroPriority.low: dict(),
            CoroPriority.normal: dict(),
            CoroPriority.high: dict()
        }
        self.priority_sequence = [
            CoroPriority.high,
            CoroPriority.normal,
            CoroPriority.low
        ]

    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        if EntityStatsMixin.StatsLevel.info == stats_level:
            func_info = full_func_info_to_printable_dict
        else:
            func_info = full_func_info_to_dict

        coroutines_set: Set = set()
        for coroutines in self.waiters.values():
            coroutines_set.update(coroutines)

        handlers_set: Set = set()
        for handlers in self.handlers.values():
            handlers_set.update(handlers)
            
        return type(self).__name__, {
            'events': {
                'low priority': {
                    'num': len(self.events[CoroPriority.low]),
                    'list': set(self.events[CoroPriority.low])
                },
                'normal priority': {
                    'num': len(self.events[CoroPriority.normal]),
                    'list': set(self.events[CoroPriority.normal])
                },
                'high priority': {
                    'num': len(self.events[CoroPriority.high]),
                    'list': set(self.events[CoroPriority.high])
                },
            },
            'waiting coroutines': {
                'num': len(coroutines_set),
                'list': coroutines_set,
            },
            'handlers': {
                'num': len(handlers_set),
                'list': [func_info(handler) for handler in handlers_set]
            }
        }

    def single_task_registration_or_immediate_processing(self, request: Optional[AsyncEventBusRequest]=None
                                                         ) -> Tuple[bool, Any, None]:
        if request is not None:
            return self.resolve_request(request)
        return True, None, None

    def full_processing_iteration(self):
        handlers = self.handlers
        waiters = self.waiters
        for priority in self.priority_sequence:
            priority_events = self.events[priority]
            if not priority_events:
                continue
            else:
                interested_events = waiters.keys() & priority_events.keys()
                for event in interested_events:
                    event_data_list: Deque = priority_events[event]
                    for i in range(len(event_data_list)):
                        event_waiters = waiters.pop(event, None)
                        if event_waiters is None:
                            break
                        else:
                            for waiter in event_waiters:
                                waiter_events = self.events_by_waiter.get(waiter, None)
                                if waiter_events:
                                    if event in waiter_events:
                                        waiter_events.remove(event)
                                        self.register_response(waiter, event_data_list.popleft())
                
                    if not event_data_list:
                        del priority_events[event]

                interested_events = handlers.keys() & priority_events.keys()
            
            if interested_events:
                processed_priority_events = set()
                handlers_info_list = list()
                for event in interested_events:
                    event_data_list = priority_events[event]
                    processed_priority_events.add(event)
                    for data in event_data_list:
                        event_handlers = handlers.get(event, None)
                        if event_handlers:
                            for handler in event_handlers:
                                handlers_info_list.append((handler, event, data))

                async def event_processing_coro(
                    interface: Interface,
                    handlers_info_list: List[Tuple[(Handler, EventID, Any)]],
                    priority: CoroPriority,
                ):
                    loop: CoroSchedulerType = interface._loop
                    current_coro_interface_buff: Interface = loop.current_coro_interface
                    ly = await agly(priority)
                    for handler, event, data in handlers_info_list:
                        try:
                            loop.current_coro_interface = None
                            handler(event, data)
                        except:
                            loop.logger.exception('AsyncEventBus. Event handler error')
                        finally:
                            loop.current_coro_interface = current_coro_interface_buff

                        await ly()

                try:
                    coro: CoroWrapperBase = self._loop.put_coro(
                        event_processing_coro,
                        handlers_info_list,
                        priority,
                    )
                    # coro.is_background_coro = True  # must not be background coro: it is not an endless coro
                except:
                    ex_type, exception, tracback = sys.exc_info()
                    raise
                finally:
                    for processed_event in processed_priority_events:
                        del priority_events[processed_event]

        self.make_dead()

    def _on_register_handler(self, event: EventID, handler: Handler):
        if event not in self.handlers:
            self.handlers[event] = set()

        self.handlers[event].add(handler)
        self.make_live()
        return True, None, None

    def _on_remove_handler(self, event: EventID, handler: Handler):
        if event in self.handlers:
            self.handlers[event].discard(handler)
            if not self.handlers[event]:
                del self.handlers[event]

        return True, None, None

    def _on_send_event(self, event: EventID, data: Any, priority: Optional[CoroPriority]=CoroPriority.low):
        events_of_priority = self.events[priority]
        if event not in events_of_priority:
            events_of_priority[event] = deque()

        events_of_priority[event].append(data)
        self.make_live()

        return True, None, None

    def _add_direct_request(self, event: EventID, data: Any, priority: Optional[CoroPriority]=CoroPriority.low) -> ValueExistence:
        self._on_send_event(event, data, priority)
        return (True, None)

    def _on_get_event(self, event: EventID, default: Any = None) -> Tuple[bool, ValueExistence, None]:
        for _, events in self.events.items():
            if event in events:
                return (True, events[event].popleft(), None)
        
        return (True, (False, default), None)

    def _on_flush_event_data(self, event: EventID):
        for _, events in self.events.items():
            if event in events:
                del events[event]
                return (True, True, None)
        
        return (True, False, None)

    def _on_wait(self, event: EventID):
        requester_id = self.current_caller_coro_info.coro_id
        requester = self.current_caller_coro_info.coro
        requester.add_on_coro_del_handler(self._on_coro_del_handler)
        if event not in self.waiters:
            self.waiters[event] = set()

        if requester_id not in self.events_by_waiter:
            self.events_by_waiter[requester_id] = set()

        self.waiters[event].add(requester_id)
        self.events_by_waiter[requester_id].add(event)
        return (False, None, None)

    def _on_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        coro_id = coro.coro_id
        if coro_id in self.events_by_waiter:
            events = self.events_by_waiter[coro_id]
            for event in events:
                if event in self.waiters:
                    if coro_id in self.waiters[event]:
                        self.waiters[event].remove(coro_id)

            del self.events_by_waiter[coro_id]
        
        return False

    def in_work(self):
        result: bool = bool(self.waiters) | bool(self.handlers)
        for priority in self.priority_sequence:
            result |= bool(self.events[priority])
        
        return self.thrifty_in_work(result)


AsyncEventBusRequest.default_service_type = AsyncEventBus


def try_send_async_event(
        backup_scheduler: Optional[CoroSchedulerType],
        event: EventID, data: Any, priority: Optional[CoroPriority]=CoroPriority.low):
    def event_sender(
            interface: Interface,
            event: EventID, data: Any, priority: Optional[CoroPriority]=CoroPriority.low):
        interface(AsyncEventBus, AsyncEventBusRequest().send_event(event, data, priority))

    try_put_coro_to(get_interface_and_loop_with_explicit_loop(backup_scheduler), event_sender, event, data, priority)
